from mindsdb.api.executor.utilities.sql import query_df
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb_sql_parser import ast
from mindsdb.integrations.utilities.date_utils import parse_local_date
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions, project_dataframe, filter_dataframe
from mindsdb.integrations.utilities.sql_utils import sort_dataframe
from mindsdb.utilities import log

from typing import Dict, List, Union
from pydantic import ValidationError

import pandas as pd

logger = log.getLogger(__name__)


class OpenBBtable(APITable):
    def _get_params_from_conditions(self, conditions: List) -> Dict:
        """Gets aggregate trade data API params from SQL WHERE conditions.

        Returns params to use for Binance API call to klines.

        Args:
            conditions (List): List of individual SQL WHERE conditions.
        """
        params: dict = {}
        # generic interpreter for conditions
        # since these are all equality conditions due to OpenBB Platform's API
        # then we can just use the first arg as the key and the second as the value
        for op, arg1, arg2 in conditions:
            if op != "=":
                raise NotImplementedError
            params[arg1] = arg2

        return params

    def _process_cols_names(self, cols: list) -> list:
        new_cols = []
        for element in cols:
            # If the element is a tuple, we want to merge the elements together
            if isinstance(element, tuple):
                # If there's more than one element we want to merge them together
                if len(element) > 1:
                    # Prevents the case where there's a multi column index and the index is a date
                    # in that instance we will have ('date', '') and this avoids having a column named 'date_'
                    new_element = "_".join(map(str, element)).rstrip("_")
                    new_cols.append(new_element)
                else:
                    new_cols.append(element[0])
            else:
                new_cols.append(element)
        return new_cols

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Selects data from the OpenBB Platform and returns it as a pandas DataFrame.

        Returns dataframe representing the OpenBB data.

        Args:
            query (ast.Select): Given SQL SELECT query
        """
        conditions = extract_comparison_conditions(query.where)
        params = self._get_params_from_conditions(conditions)

        try:
            if params is None:
                logger.error("At least cmd needs to be added!")
                raise Exception("At least cmd needs to be added!")

            # Get the OpenBB command to get the data from
            cmd = params.pop("cmd")

            # Ensure that the cmd provided is a valid OpenBB command
            available_cmds = [f"obb{cmd}" for cmd in list(self.handler.obb.coverage.commands.keys())]
            if cmd not in available_cmds:
                logger.error(f"The command provided is not supported by OpenBB! Choose one of the following: {', '.join(available_cmds)}")
                raise Exception(f"The command provided is not supported by OpenBB! Choose one of the following: {', '.join(available_cmds)}")

            args = ""
            # If there are parameters create arguments as a string
            if params:
                for arg, val in params.items():
                    args += f"{arg}={val},"

                # Remove the additional ',' added at the end
                if args:
                    args = args[:-1]

            # Recreate the OpenBB command with the arguments
            openbb_cmd = f"self.handler.{cmd}({args})"

            # Execute the OpenBB command and return the OBBject
            openbb_object = eval(openbb_cmd)

            # Transform the OBBject into a pandas DataFrame
            data = openbb_object.to_df()

            # Check if index is a datetime, if it is we want that as a column
            if isinstance(data.index, pd.DatetimeIndex):
                data.reset_index(inplace=True)

            # Process column names
            data.columns = self._process_cols_names(data.columns)

        except Exception as e:
            logger.error(f"Error accessing data from OpenBB: {e}!")
            raise Exception(f"Error accessing data from OpenBB: {e}!")

        return data


def create_table_class(
    params_metadata,
    response_metadata,
    obb_function,
    func_docs="",
    provider=None
):
    """Creates a table class for the given OpenBB Platform function."""
    mandatory_fields = [key for key in params_metadata['fields'].keys() if params_metadata['fields'][key].is_required() is True]
    response_columns = list(response_metadata['fields'].keys())

    class AnyTable(APITable):
        def _get_params_from_conditions(self, conditions: List) -> Dict:
            """Gets aggregate trade data API params from SQL WHERE conditions.

            Returns params to use for Binance API call to klines.

            Args:
                conditions (List): List of individual SQL WHERE conditions.
            """
            params: dict = {}
            # generic interpreter for conditions
            # since these are all equality conditions due to OpenBB Platform's API
            # then we can just use the first arg as the key and the second as the value
            for op, arg1, arg2 in conditions:
                if op == "=":
                    params[arg1] = arg2

            return params

        def select(self, query: ast.Select) -> pd.DataFrame:
            """Selects data from the OpenBB Platform and returns it as a pandas DataFrame.

            Returns dataframe representing the OpenBB data.

            Args:
                query (ast.Select): Given SQL SELECT query
            """
            conditions = extract_comparison_conditions(query.where)
            arg_params = self._get_params_from_conditions(conditions=conditions)

            params = {}
            if provider is not None:
                params['provider'] = provider

            filters = []
            mandatory_args_set = {key: False for key in mandatory_fields}
            columns_to_add = {}
            strict_filter = arg_params.get('strict_filter', False)

            for op, arg1, arg2 in conditions:
                if op == 'or':
                    raise NotImplementedError('OR is not supported')

                if arg1 in mandatory_fields:
                    mandatory_args_set[arg1] = True

                if ('start_' + arg1 in params_metadata['fields'] and arg1 in response_columns and arg2 is not None):

                    if response_metadata['fields'][arg1].annotation == 'datetime':
                        date = parse_local_date(arg2)
                        interval = arg_params.get('interval', '1d')

                        if op == '>':
                            params['start_' + arg1] = date.strftime('%Y-%m-%d')
                        elif op == '<':
                            params['end_' + arg1] = date.strftime('%Y-%m-%d')
                        elif op == '>=':
                            date = date - pd.Timedelta(interval)
                            params['start_' + arg1] = date.strftime('%Y-%m-%d')
                        elif op == '<=':
                            date = date + pd.Timedelta(interval)
                            params['end_' + arg1] = date.strftime('%Y-%m-%d')
                        elif op == '=':
                            date = date - pd.Timedelta(interval)
                            params['start_' + arg1] = date.strftime('%Y-%m-%d')
                            date = date + pd.Timedelta(interval)
                            params['end_' + arg1] = date.strftime('%Y-%m-%d')

                elif arg1 in params_metadata['fields'] or not strict_filter:
                    if op == '=':
                        params[arg1] = arg2
                        columns_to_add[arg1] = arg2

                filters.append([op, arg1, arg2])

            if not all(mandatory_args_set.values()):
                missing_args = ", ".join([k for k, v in mandatory_args_set.items() if v is False])
                text = f"You must specify the following arguments in the WHERE statement: {missing_args}\n"

                # Create docstring for the current function
                text += "\nDocstring:"
                for param in params_metadata['fields']:
                    field = params_metadata['fields'][param]
                    if getattr(field.annotation, '__origin__', None) is Union:
                        annotation = f"Union[{', '.join(arg.__name__ for arg in field.annotation.__args__)}]"
                    else:
                        annotation = field.annotation.__name__
                    text += f"\n  * {param}{'' if field.is_required() else ' (optional)'}: {annotation}\n{field.description}"

                text += f"\n\nFor more information check {func_docs}"

                raise NotImplementedError(text)

            try:
                # Handle limit keyword correctly since it can't be parsed as a WHERE arg (i.e. WHERE limit = 50)
                if query.limit is not None and 'limit' in params_metadata['fields']:
                    params['limit'] = query.limit.value
                obbject = obb_function(**params)

                # Extract data in dataframe format
                result = obbject.to_df()

                if result is None:
                    raise Exception(f"For more information check {func_docs}.")

                # Check if index is a datetime, if it is we want that as a column
                if isinstance(result.index, pd.DatetimeIndex):
                    result.reset_index(inplace=True)

                if query.order_by:
                    result = sort_dataframe(result, query.order_by)

                if query.limit is not None:
                    result = result.head(query.limit.value)

                    if result is None:
                        raise Exception(f"For more information check {func_docs}.")

                for key in columns_to_add:
                    result[key] = params[key]

                # filter targets
                result = filter_dataframe(result, filters)

                if result is None:
                    raise Exception(f"For more information check {func_docs}.")

                columns = self.get_columns()

                columns += [col for col in result.columns if col not in columns]

                for full_target in query.targets:
                    if isinstance(full_target, ast.Star):
                        continue
                    if isinstance(full_target, ast.Identifier):
                        target = full_target.parts[-1].lower()
                    elif isinstance(full_target, ast.Function):
                        target = full_target.args[0].parts[-1].lower()
                    else:
                        # Could be a window function or other operation we can't handle. Defer to DuckDB.
                        return query_df(result, query)
                    if target not in columns:
                        raise ValueError(f"Unknown column '{target}' in 'field list'")

                # project targets
                try:
                    result = project_dataframe(result, query.targets, columns)
                except NotImplementedError:
                    # Target contains a function that we need DuckDB to resolve.
                    return query_df(result, query)
                return result

            except AttributeError as e:
                logger.info(f'Encountered error while executing OpenBB select: {str(e)}')

                # Create docstring for the current function
                text = "Docstring:"
                for param in params_metadata['fields']:
                    field = params_metadata['fields'][param]
                    if getattr(field.annotation, '__origin__', None) is Union:
                        annotation = f"Union[{', '.join(arg.__name__ for arg in field.annotation.__args__)}]"
                    else:
                        annotation = field.annotation.__name__
                    text += f"\n  * {param}{'' if field.is_required() else ' (optional)'}: {annotation}\n{field.description}"

                text += f"\n\nFor more information check {func_docs}"

                raise Exception(f"{str(e)}\n\n{text}.") from e

            except ValidationError as e:
                logger.info(f'Encountered error while executing OpenBB select: {str(e)}')

                # Create docstring for the current function
                text = "Docstring:"
                for param in params_metadata['fields']:
                    field = params_metadata['fields'][param]
                    if getattr(field.annotation, '__origin__', None) is Union:
                        annotation = f"Union[{', '.join(arg.__name__ for arg in field.annotation.__args__)}]"
                    else:
                        annotation = field.annotation.__name__
                    text += f"\n  * {param}{'' if field.is_required() else ' (optional)'}: {annotation}\n{field.description}"

                text += f"\n\nFor more information check {func_docs}"

                raise Exception(f"{str(e)}\n\n{text}.") from e

            except Exception as e:
                logger.info(f'Encountered error while executing OpenBB select: {str(e)}')

                #  TODO: This one doesn't work because it's taken care of from MindsDB side
                if "Table not found" in str(e):
                    raise Exception(f"{str(e)}\n\nCheck if the method exists here: {func_docs}.\n\n  -  If it doesn't you may need to look for the parent module to check whether there's a typo in the naming.\n- If it does you may need to install a new extension to the OpenBB Platform, and you can see what is available at https://my.openbb.co/app/platform/extensions.") from e

                if "Missing credential" in str(e):
                    raise Exception(f"{str(e)}\n\nGo to https://my.openbb.co/app/platform/api-keys to set this API key, for free.") from e

                # Catch all other errors
                # Create docstring for the current function
                text = "Docstring:"
                for param in params_metadata['fields']:
                    field = params_metadata['fields'][param]
                    if getattr(field.annotation, '__origin__', None) is Union:
                        annotation = f"Union[{', '.join(arg.__name__ for arg in field.annotation.__args__)}]"
                    else:
                        annotation = field.annotation.__name__
                    text += f"\n  * {param}{'' if field.is_required() else ' (optional)'}: {annotation}\n{field.description}"

                text += f"\n\nFor more information check {func_docs}"

                raise Exception(f"{str(e)}\n\n{text}.") from e

        def get_columns(self):
            return response_columns

    return AnyTable



import os
import typing

from opentelemetry import trace  # noqa: F401
from opentelemetry import metrics  # noqa: F401
from opentelemetry.exporter.otlp.proto.grpc._log_exporter import OTLPLogExporter as OTLPLogExporterGRPC
from opentelemetry.exporter.otlp.proto.http._log_exporter import OTLPLogExporter as OTLPLogExporterHTTP
from opentelemetry.sdk._logs._internal.export import LogExporter
from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import OTLPMetricExporter as OTLPMetricExporterGRPC
from opentelemetry.exporter.otlp.proto.http.metric_exporter import OTLPMetricExporter as OTLPMetricExporterHTTP
from opentelemetry.sdk.metrics.export import MetricExporter, ConsoleMetricExporter
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter as OTLPSpanExporterGRPC
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter as OTLPSpanExporterHTTP
from opentelemetry.sdk.trace.export import SpanExporter, ConsoleSpanExporter
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace.sampling import TraceIdRatioBased

from mindsdb.utilities.otel.logger import setup_logger
from mindsdb.utilities.otel.meter import setup_meter
from mindsdb.utilities.otel.tracer import setup_tracer
from mindsdb.utilities.utils import parse_csv_attributes
from mindsdb.utilities import log

logger = log.getLogger(__name__)

# Check OpenTelemetry exporter type
OTEL_EXPORTER_TYPE = os.getenv("OTEL_EXPORTER_TYPE", "console")  # console or otlp

# Define OpenTelemetry exporter protocol
OTEL_EXPORTER_PROTOCOL = os.getenv("OTEL_EXPORTER_PROTOCOL", "grpc")  # grpc or http

# Define OTLP endpoint. If not set, the default OTLP endpoint will be used
OTEL_OTLP_ENDPOINT = os.getenv("OTEL_OTLP_ENDPOINT", "http://localhost:4317")

# Define OTLP logging endpoint. If not set, the default OTLP logging endpoint will be used
OTEL_OTLP_LOGGING_ENDPOINT = os.getenv("OTEL_OTLP_LOGGING_ENDPOINT", OTEL_OTLP_ENDPOINT)

# Define OTLP tracing endpoint. If not set, the default OTLP tracing endpoint will be used
OTEL_OTLP_TRACING_ENDPOINT = os.getenv("OTEL_OTLP_TRACING_ENDPOINT", OTEL_OTLP_ENDPOINT)

# Define OTLP metrics endpoint. If not set, the default OTLP metrics endpoint will be used
OTEL_OTLP_METRICS_ENDPOINT = os.getenv("OTEL_OTLP_METRICS_ENDPOINT", OTEL_OTLP_ENDPOINT)

# Define service name
OTEL_SERVICE_NAME = os.getenv("OTEL_SERVICE_NAME", "mindsdb")

# Define service instace ID
OTEL_SERVICE_INSTANCE_ID = os.getenv("OTEL_SERVICE_INSTANCE_ID", "mindsdb-instance")

# The name of the environment we"re on, by default local for development, this is set differently per-env in our Helm
# chart values files
OTEL_SERVICE_ENVIRONMENT = os.getenv("OTEL_SERVICE_ENVIRONMENT", "local").lower()

# Define service release
OTEL_SERVICE_RELEASE = os.getenv("OTEL_SERVICE_RELEASE", "local").lower()

# Define how often to capture traces
OTEL_TRACE_SAMPLE_RATE = float(os.getenv("OTEL_TRACE_SAMPLE_RATE", "1.0"))

# Define extra attributes
OTEL_EXTRA_ATTRIBUTES = os.getenv("OTEL_EXTRA_ATTRIBUTES", "")

# By default, we have Open Telemetry SDK enabled on all envs, except for local which is disabled by default.
OTEL_SDK_DISABLED = (os.getenv("OTEL_SDK_DISABLED", "false").lower() == "true"
                     or os.getenv("OTEL_SERVICE_ENVIRONMENT", "local").lower() == "local")

# Define if OpenTelemetry logging is disabled. By default, it is disabled.
OTEL_LOGGING_DISABLED = os.getenv("OTEL_LOGGING_DISABLED", "true").lower() == "true"

# Define if OpenTelemetry tracing is disabled. By default, it is enabled.
OTEL_TRACING_DISABLED = os.getenv("OTEL_TRACING_DISABLED", "false").lower() == "true"

# Define if OpenTelemetry metrics is disabled. By default, it is disabled.
OTEL_METRICS_DISABLED = os.getenv("OTEL_METRICS_DISABLED", "true").lower() == "true"

# If you want to enable Open Telemetry on local for some reason please set OTEL_SDK_FORCE_RUN to true
OTEL_SDK_FORCE_RUN = os.getenv("OTEL_SDK_FORCE_RUN", "false").lower() == "true"


def get_otel_attributes() -> dict:
    """
    Get OpenTelemetry attributes

    Returns:
        dict: OpenTelemetry attributes
    """

    base_attributes = {
        "service.name": OTEL_SERVICE_NAME,
        "service.instance.id": OTEL_SERVICE_INSTANCE_ID,
        "environment": OTEL_SERVICE_ENVIRONMENT,
        "release": OTEL_SERVICE_RELEASE,
    }

    extra_attributes = {}
    try:
        extra_attributes = parse_csv_attributes(OTEL_EXTRA_ATTRIBUTES)
    except Exception as e:
        logger.error(f"Failed to parse OTEL_EXTRA_ATTRIBUTES: {e}")

    attributes = {**extra_attributes, **base_attributes}  # Base attributes take precedence over extra attributes

    return attributes


def get_logging_exporter() -> typing.Optional[LogExporter]:
    """
    Get OpenTelemetry logging exporter.

    Returns:
        OTLPLogExporter: OpenTelemetry logging exporter
    """

    if OTEL_EXPORTER_TYPE == "otlp":

        if OTEL_EXPORTER_PROTOCOL == "grpc":
            return OTLPLogExporterGRPC(
                endpoint=OTEL_OTLP_LOGGING_ENDPOINT,
                insecure=True
            )

        elif OTEL_EXPORTER_PROTOCOL == "http":
            return OTLPLogExporterHTTP(
                endpoint=OTEL_OTLP_LOGGING_ENDPOINT
            )

    return None


def get_span_exporter() -> SpanExporter:
    """
    Get OpenTelemetry span exporter

    Returns:
        OTLPSpanExporter: OpenTelemetry span exporter
    """

    if OTEL_EXPORTER_TYPE == "otlp":

        if OTEL_EXPORTER_PROTOCOL == "grpc":
            return OTLPSpanExporterGRPC(
                endpoint=OTEL_OTLP_TRACING_ENDPOINT,
                insecure=True
            )

        elif OTEL_EXPORTER_PROTOCOL == "http":
            return OTLPSpanExporterHTTP(
                endpoint=OTEL_OTLP_TRACING_ENDPOINT
            )

    return ConsoleSpanExporter()


def get_metrics_exporter() -> typing.Optional[MetricExporter]:
    """
    Get OpenTelemetry metrics exporter

    Returns:
        OTLPLogExporter: OpenTelemetry metrics exporter
    """

    if OTEL_EXPORTER_TYPE == "otlp":

        if OTEL_EXPORTER_PROTOCOL == "grpc":
            return OTLPMetricExporterGRPC(
                endpoint=OTEL_OTLP_METRICS_ENDPOINT,
                insecure=True
            )

        elif OTEL_EXPORTER_PROTOCOL == "http":
            return OTLPMetricExporterHTTP(
                endpoint=OTEL_OTLP_METRICS_ENDPOINT
            )

    return ConsoleMetricExporter()


if not OTEL_SDK_DISABLED or OTEL_SDK_FORCE_RUN:
    logger.info("OpenTelemetry enabled")
    logger.info(f"OpenTelemetry exporter type: {OTEL_EXPORTER_TYPE}")
    logger.info(f"OpenTelemetry service name: {OTEL_SERVICE_NAME}")
    logger.info(f"OpenTelemetry service environment: {OTEL_SERVICE_ENVIRONMENT}")
    logger.info(f"OpenTelemetry service release: {OTEL_SERVICE_RELEASE}")
    logger.info(f"OpenTelemetry trace sample rate: {OTEL_TRACE_SAMPLE_RATE}")
    logger.info(f"OpenTelemetry extra attributes: {OTEL_EXTRA_ATTRIBUTES}")

    # Define OpenTelemetry resources (e.g., service name)
    attributes = get_otel_attributes()

    # Define OpenTelemetry sampler
    sampler = TraceIdRatioBased(OTEL_TRACE_SAMPLE_RATE)

    # Define OpenTelemetry resources (e.g., service name)
    resource = Resource(attributes=attributes)

    if not OTEL_LOGGING_DISABLED:
        logger.info("OpenTelemetry Logging is enabled")
        setup_logger(resource, get_logging_exporter())

    if not OTEL_TRACING_DISABLED:
        logger.info("OpenTelemetry Tracing is enabled")
        setup_tracer(resource, sampler, get_span_exporter())

    if not OTEL_METRICS_DISABLED:
        logger.info("OpenTelemetry Metrics is enabled")
        setup_meter(resource, get_metrics_exporter())



import pandas as pd
from typing import Text, List
from mindsdb_sql_parser import ast
from abc import ABC, abstractmethod
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions, filter_dataframe


class BaseQueryParser(ABC):
    """
    Parses a SQL query into its component parts.

    Parameters
    ----------
    query : ast
        Given SQL query.
    """

    def __init__(self, query: ast):
        self.query = query

    @abstractmethod
    def parse_query(self):
        """
        Parses a SQL statement into its components.
        """
        pass

    def parse_where_clause(self) -> List[List[Text]]:
        """
        Parses the WHERE clause of the query.
        """
        where_conditions = extract_comparison_conditions(self.query.where)
        return where_conditions


class BaseQueryExecutor():
    """
    Executes a SQL query.

    Parameters
    ----------
    query : ast
        Given SQL query.
    """

    def __init__(self, df: pd.DataFrame, where_conditions: List[List[Text]]):
        self.df = df
        self.where_conditions = where_conditions

    def execute_query(self):
        """
        Executes the SQL query.
        """
        self.execute_where_clause()

        return self.df

    def execute_where_clause(self):
        """
        Executes the where clause of the query.
        """
        if len(self.where_conditions) > 0:
            self.df = filter_dataframe(self.df, self.where_conditions)



from typing import Text, Dict, Callable

from botbuilder.schema import Activity, ActivityTypes
from botbuilder.schema import ChannelAccount
from botframework.connector import ConnectorClient
from botframework.connector.auth import MicrosoftAppCredentials
from mindsdb.utilities import log
from mindsdb_sql_parser import parse_sql

from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.integrations.libs.api_handler import APIChatHandler
from mindsdb.interfaces.chatbot.types import ChatBotMessage

logger = log.getLogger(__name__)


class MSTeamsHandler(APIChatHandler):
    """
    The Microsoft Teams handler implementation.
    """

    name = 'teams'

    def __init__(self, name: str, **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

        self.service_url = None
        self.channel_id = None
        self.bot_id = None
        self.conversation_id = None

    def connect(self) -> MicrosoftAppCredentials:
        """
        Set up the connection required by the handler.

        Returns
        -------
        MicrosoftAppCredentials
            Client object for interacting with the Microsoft Teams app.
        """
        if self.is_connected:
            return self.connection

        self.connection = MicrosoftAppCredentials(
            app_id=self.connection_data["client_id"],
            password=self.connection_data["client_secret"]
        )

        self.is_connected = True

        return self.connection

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.

        Returns
        -------
        StatusResponse
            Response object with the status of the connection.
        """
        response = StatusResponse(False)

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to Microsoft Teams: {e}!')
            response.success = False
            response.error_message = str(e)

        self.is_connected = response.success

        return response

    def native_query(self, query: Text) -> StatusResponse:
        """
        Receive and process a raw query.

        Parameters
        ----------
        query: Text
            Query in the native format.

        Returns
        -------
        StatusResponse
            Response object with the result of the query.
        """
        ast = parse_sql(query)

        return self.query(ast)
    
    def get_chat_config(self) -> Dict:
        """
        Get the configuration for the chatbot.
        This method is required for the implementation of the chatbot.

        Returns
        -------
        Dict
            Configuration for the chatbot.
        """
        params = {
            'polling': {
                'type': 'webhook'
            }
        }

        return params
    
    def get_my_user_name(self) -> Text:
        """
        Get the name of the signed in user.
        This method is required for the implementation of the chatbot.

        Returns
        -------
        Text
            Name of the signed in user.
        """
        return None
    
    def on_webhook(self, request: Dict, callback: Callable) -> None:
        """
        Handle a webhook request.

        Parameters
        ----------
        request: Dict
            The incoming webhook request.

        callback: Callable
            Callback function to call after parsing the request.
        """
        self.service_url = request["serviceUrl"]
        self.channel_id = request["channelId"]
        self.bot_id = request["from"]["id"]
        self.conversation_id = request["conversation"]["id"]

        chat_bot_message = ChatBotMessage(
            ChatBotMessage.Type.DIRECT,
            text=request["text"],
            user=request["from"]["id"],
            destination=request["recipient"]["id"]
        )

        callback(
            chat_id=request['conversation']['id'],
            message=chat_bot_message
        )    
            
    def respond(self, message: ChatBotMessage) -> None:
        """
        Send a response to the chatbot.

        Parameters
        ----------
        message: ChatBotMessage
            The message to send.
        """
        credentials = self.connect()

        connector = ConnectorClient(credentials, base_url=self.service_url)
        connector.conversations.send_to_conversation(
            self.conversation_id,
            Activity(
                type=ActivityTypes.message,
                channel_id=self.channel_id,
                recipient=ChannelAccount(
                    id=message.destination
                ),
                from_property=ChannelAccount(
                    id=self.bot_id
                ),
                text=message.text
            )
        )



from mindsdb.integrations.libs.const import HANDLER_TYPE
from .__about__ import __version__ as version, __description__ as description

try:
    from .google_analytics_handler import GoogleAnalyticsHandler as Handler

    import_error = None
except Exception as e:
    Handler = None
    import_error = e
title = 'Google Analytics'
name = 'google_analytics'
type = HANDLER_TYPE.DATA
icon_path = 'icon.svg'
permanent = False
__all__ = [
    'Handler', 'version', 'name', 'type', 'title', 'description', 'import_error', 'icon_path'
]



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    database={
        "type": ARG_TYPE.STR,
        "description": "The database file where the data will be stored.",
    },
    sync_url={
        "type": ARG_TYPE.STR,
        "description": "The database URL where the data is synced with.",
    },
    auth_token={
        "type": ARG_TYPE.STR,
        "description": "The JWT auth token to authenticate with the sync database.",
        "secret": True
    },
)

connection_args_example = OrderedDict(database="chinook.db")




import re
import csv
import inspect
from io import StringIO
from typing import Iterable, List, Optional, Any

import pandas as pd
from mindsdb_sql_parser import parse_sql
from mindsdb_sql_parser.ast import Select, Show, Describe, Explain, Identifier

from mindsdb.utilities import log
from mindsdb.utilities.context import context as ctx
from mindsdb.integrations.utilities.query_traversal import query_traversal

logger = log.getLogger(__name__)


def list_to_csv_str(array: List[List[Any]]) -> str:
    """Convert a 2D array into a CSV string.

    Args:
        array (List[List[Any]]): A 2D array/list of values to convert to CSV format

    Returns:
        str: The array formatted as a CSV string using Excel dialect
    """
    output = StringIO()
    writer = csv.writer(output, dialect='excel')
    str_array = [[str(item) for item in row] for row in array]
    writer.writerows(str_array)
    return output.getvalue()


def split_table_name(table_name: str) -> List[str]:
    """Split table name from llm to parst

    Args:
        table_name (str): input table name

    Returns:
        List[str]: parts of table identifier like ['database', 'schema', 'table']

    Example:
        'input': '`aaa`.`bbb.ccc`', 'output': ['aaa', 'bbb.ccc']
        'input': '`aaa`.`bbb`.`ccc`', 'output': ['aaa', 'bbb', 'ccc']
        'input': 'aaa.bbb', 'output': ['aaa', 'bbb']
        'input': '`aaa.bbb`', 'output': ['aaa.bbb']
        'input': '`aaa.bbb.ccc`', 'output': ['aaa.bbb.ccc']
        'input': 'aaa.`bbb`', 'output': ['aaa', 'bbb']
        'input': 'aaa.bbb.ccc', 'output': ['aaa', 'bbb', 'ccc']
        'input': 'aaa.`bbb.ccc`', 'output': ['aaa', 'bbb.ccc']
        'input': '`aaa`.`bbb.ccc`', 'output': ['aaa', 'bbb.ccc']
    """
    result = []
    current = ''
    in_backticks = False

    i = 0
    while i < len(table_name):
        if table_name[i] == '`':
            in_backticks = not in_backticks
        elif table_name[i] == '.' and not in_backticks:
            if current:
                result.append(current.strip('`'))
                current = ''
        else:
            current += table_name[i]
        i += 1

    if current:
        result.append(current.strip('`'))

    return result


class SQLAgent:
    def __init__(
            self,
            command_executor,
            databases: List[str],
            databases_struct: dict,
            include_tables: Optional[List[str]] = None,
            ignore_tables: Optional[List[str]] = None,
            sample_rows_in_table_info: int = 3,
            cache: Optional[dict] = None
    ):
        self._command_executor = command_executor
        self._mindsdb_db_struct = databases_struct

        self._sample_rows_in_table_info = int(sample_rows_in_table_info)

        self._tables_to_include = include_tables
        self._tables_to_ignore = []
        self._databases = databases
        if not self._tables_to_include:
            # ignore_tables and include_tables should not be used together.
            # include_tables takes priority if it's set.
            self._tables_to_ignore = ignore_tables or []
        self._cache = cache

    def _call_engine(self, query: str, database=None):
        # switch database
        ast_query = parse_sql(query.strip('`'))
        self._check_permissions(ast_query)

        if database is None:
            # if we use tables with prefixes it should work for any database
            database = self._databases[0]

        ret = self._command_executor.execute_command(
            ast_query,
            database_name=database
        )
        return ret

    def _check_permissions(self, ast_query):
        # check type of query
        if not isinstance(ast_query, (Select, Show, Describe, Explain)):
            raise ValueError(f"Query is not allowed: {ast_query.to_string()}")

        # Check tables
        if self._tables_to_include:
            tables_parts = [split_table_name(x) for x in self._tables_to_include]
            no_schema_parts = []
            for t in tables_parts:
                if len(t) == 3:
                    no_schema_parts.append([t[0], t[2]])
            tables_parts += no_schema_parts

            def _check_f(node, is_table=None, **kwargs):
                if is_table and isinstance(node, Identifier):
                    if node.parts not in tables_parts:
                        raise ValueError(f"Table {'.'.join(node.parts)} not found. Available tables: {', '.join(self._tables_to_include)}")

            query_traversal(ast_query, _check_f)

    def get_usable_table_names(self) -> Iterable[str]:
        """Get a list of tables that the agent has access to.

        Returns:
            Iterable[str]: list with table names
        """
        cache_key = f'{ctx.company_id}_{",".join(self._databases)}_tables'

        # first check cache and return if found
        if self._cache:
            cached_tables = self._cache.get(cache_key)
            if cached_tables:
                return cached_tables

        if self._tables_to_include:
            return self._tables_to_include

        result_tables = []

        for db_name in self._mindsdb_db_struct:
            handler = self._command_executor.session.integration_controller.get_data_handler(db_name)

            schemas_names = list(self._mindsdb_db_struct[db_name].keys())
            if len(schemas_names) > 1 and None in schemas_names:
                raise Exception('default schema and named schemas can not be used in same filter')

            if None in schemas_names:
                # get tables only from default schema
                response = handler.get_tables()
                tables_in_default_schema = list(response.data_frame.table_name)
                schema_tables_restrictions = self._mindsdb_db_struct[db_name][None]     # None - is default schema
                if schema_tables_restrictions is None:
                    for table_name in tables_in_default_schema:
                        result_tables.append([db_name, table_name])
                else:
                    for table_name in schema_tables_restrictions:
                        if table_name in tables_in_default_schema:
                            result_tables.append([db_name, table_name])
            else:
                if 'all' in inspect.signature(handler.get_tables).parameters:
                    response = handler.get_tables(all=True)
                else:
                    response = handler.get_tables()
                response_schema_names = list(response.data_frame.table_schema.unique())
                schemas_intersection = set(schemas_names) & set(response_schema_names)
                if len(schemas_intersection) == 0:
                    raise Exception('There are no allowed schemas in ds')

                for schema_name in schemas_intersection:
                    schema_sub_df = response.data_frame[response.data_frame['table_schema'] == schema_name]
                    if self._mindsdb_db_struct[db_name][schema_name] is None:
                        # all tables from schema allowed
                        for row in schema_sub_df:
                            result_tables.append([db_name, schema_name, row['table_name']])
                    else:
                        for table_name in self._mindsdb_db_struct[db_name][schema_name]:
                            if table_name in schema_sub_df['table_name'].values:
                                result_tables.append([db_name, schema_name, table_name])

        result_tables = ['.'.join(x) for x in result_tables]
        if self._cache:
            self._cache.set(cache_key, set(result_tables))
        return result_tables

    def _resolve_table_names(self, table_names: List[str], all_tables: List[Identifier]) -> List[Identifier]:
        """
        Tries to find table (which comes directly from an LLM) by its name
        Handles backticks (`) and tables without databases
        """

        # index to lookup table
        tables_idx = {}
        for table in all_tables:
            # by name
            if len(table.parts) == 3:
                tables_idx[tuple(table.parts[1:])] = table
            else:
                tables_idx[(table.parts[-1],)] = table
            # by path
            tables_idx[tuple(table.parts)] = table

        tables = []
        for table_name in table_names:
            if not table_name.strip():
                continue

            # Some LLMs (e.g. gpt-4o) may include backticks or quotes when invoking tools.
            table_parts = split_table_name(table_name)
            if len(table_parts) == 1:
                # most likely LLM enclosed all table name in backticks `database.table`
                table_parts = split_table_name(table_name)

            # resolved table
            table_identifier = tables_idx.get(tuple(table_parts))

            if table_identifier is None:
                raise ValueError(f"Table {table} not found in the database")
            tables.append(table_identifier)

        return tables

    def get_table_info(self, table_names: Optional[List[str]] = None) -> str:
        """ Get information about specified tables.
        Follows best practices as specified in: Rajkumar et al, 2022 (https://arxiv.org/abs/2204.00498)
        If `sample_rows_in_table_info`, the specified number of sample rows will be
        appended to each table description. This can increase performance as demonstrated in the paper.
        """

        all_tables = [Identifier(name) for name in self.get_usable_table_names()]

        if table_names is not None:
            all_tables = self._resolve_table_names(table_names, all_tables)

        tables_info = []
        for table in all_tables:
            key = f"{ctx.company_id}_{table}_info"
            table_info = self._cache.get(key) if self._cache else None
            if table_info is None:
                table_info = self._get_single_table_info(table)
                if self._cache:
                    self._cache.set(key, table_info)

            tables_info.append(table_info)

        return "\n\n".join(tables_info)

    def _get_single_table_info(self, table: Identifier) -> str:
        if len(table.parts) < 2:
            raise ValueError(f"Database is required for table: {table}")
        if len(table.parts) == 3:
            integration, schema_name, table_name = table.parts[-3:]
        else:
            schema_name = None
            integration, table_name = table.parts[-2:]

        table_str = str(table)

        dn = self._command_executor.session.datahub.get(integration)

        fields, dtypes = [], []
        for column in dn.get_table_columns(table_name, schema_name):
            fields.append(column['name'])
            dtypes.append(column.get('type', ''))

        info = f'Table named `{table_str}`:\n'
        info += f"\nSample with first {self._sample_rows_in_table_info} rows from table {table_str} in CSV format (dialect is 'excel'):\n"
        info += self._get_sample_rows(table_str, fields) + "\n"
        info += '\nColumn data types: ' + ",\t".join(
            [f'\n`{field}` : `{dtype}`' for field, dtype in zip(fields, dtypes)]) + '\n'  # noqa
        return info

    def _get_sample_rows(self, table: str, fields: List[str]) -> str:
        command = f"select {', '.join(fields)} from {table} limit {self._sample_rows_in_table_info};"
        try:
            ret = self._call_engine(command)
            sample_rows = ret.data.to_lists()

            def truncate_value(val):
                str_val = str(val)
                return str_val if len(str_val) < 100 else (str_val[:100] + '...')

            sample_rows = list(
                map(lambda row: [truncate_value(value) for value in row], sample_rows))
            sample_rows_str = "\n" + list_to_csv_str([fields] + sample_rows)
        except Exception as e:
            logger.warning(e)
            sample_rows_str = "\n" + "\t [error] Couldn't retrieve sample rows!"

        return sample_rows_str

    def _clean_query(self, query: str) -> str:
        # Sometimes LLM can input markdown into query tools.
        cmd = re.sub(r'```(sql)?', '', query)
        return cmd

    def query(self, command: str, fetch: str = "all") -> str:
        """Execute a SQL command and return a string representing the results.
        If the statement returns rows, a string of the results is returned.
        If the statement returns no rows, an empty string is returned.
        """

        def _repr_result(ret):
            limit_rows = 30

            columns_str = ', '.join([repr(col.name) for col in ret.columns])
            res = f'Output columns: {columns_str}\n'

            data = ret.to_lists()
            if len(data) > limit_rows:
                df = pd.DataFrame(data, columns=[col.name for col in ret.columns])

                res += f'Result has {len(data)} rows. Description of data:\n'
                res += str(df.describe(include='all')) + '\n\n'
                res += f'First {limit_rows} rows:\n'

            else:
                res += "Result in CSV format (dialect is 'excel'):\n"
            res += list_to_csv_str([[col.name for col in ret.columns]] + data[:limit_rows])
            return res

        ret = self._call_engine(self._clean_query(command))
        if fetch == "all":
            result = _repr_result(ret.data)
        elif fetch == "one":
            result = "Result in CSV format (dialect is 'excel'):\n"
            result += list_to_csv_str([[col.name for col in ret.data.columns]] + [ret.data.to_lists()[0]])
        else:
            raise ValueError("Fetch parameter must be either 'one' or 'all'")
        return str(result)

    def get_table_info_safe(self, table_names: Optional[List[str]] = None) -> str:
        try:
            return self.get_table_info(table_names)
        except Exception as e:
            return f"Error: {e}"

    def query_safe(self, command: str, fetch: str = "all") -> str:
        try:
            return self.query(command, fetch)
        except Exception as e:
            msg = f"Error: {e}"
            if 'does not exist' in msg and ' relation ' in msg:
                msg += '\nAvailable tables: ' + ', '.join(self.get_usable_table_names())
            return msg



from mindsdb.integrations.libs.const import HANDLER_TYPE

from .file_handler import FileHandler as Handler
from .__about__ import __version__ as version


title = 'File'
name = 'files'
type = HANDLER_TYPE.DATA
icon_path = "icon.svg"

permanent = True

__all__ = [
    'Handler', 'version', 'name', 'type', 'title', 'icon_path'
]



from mindsdb_sql_parser.ast import Select, BinaryOperation, Identifier, Constant, Star
from mindsdb.integrations.libs.vectordatabase_handler import TableField

from typing import Any, List, Optional

from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.vectorstores import VectorStore


class MDBVectorStore(VectorStore):

    def __init__(self, kb_table) -> None:
        self.kb_table = kb_table

    @property
    def embeddings(self) -> Optional[Embeddings]:
        return None

    def similarity_search(
        self,
        query: str,
        k: int = 4,
        **kwargs: Any,
    ) -> List[Document]:

        query = Select(
            targets=[Star()],
            where=BinaryOperation(op='=', args=[
                Identifier(TableField.CONTENT.value), Constant(query)
            ]),
            limit=Constant(k),
        )

        df = self.kb_table.select_query(query)

        docs = []
        for _, row in df.iterrows():
            metadata = row[TableField.METADATA.value]
            if metadata is None:
                metadata = {}
            docs.append(Document(
                page_content=row[TableField.CONTENT.value],
                metadata=metadata
            ))

        return docs

    def add_texts(self, *args, **kwargs) -> List[str]:
        raise NotImplementedError

    @classmethod
    def from_texts(self, *args, **kwargs):
        raise NotImplementedError



import datetime
from typing import Dict, List, Optional

from sqlalchemy import null
from sqlalchemy.orm.attributes import flag_modified

from mindsdb.interfaces.storage import db
from mindsdb.interfaces.database.projects import ProjectController


class SkillsController:
    '''Handles CRUD operations at the database level for Skills'''

    def __init__(self, project_controller: ProjectController = None):
        if project_controller is None:
            project_controller = ProjectController()
        self.project_controller = project_controller

    def get_skill(self, skill_name: str, project_name: str = 'mindsdb') -> Optional[db.Skills]:
        '''
        Gets a skill by name. Skills are expected to have unique names.

        Parameters:
            skill_name (str): The name of the skill
            project_name (str): The name of the containing project

        Returns:
            skill (Optional[db.Skills]): The database skill object

        Raises:
            ValueError: If `project_name` does not exist
        '''

        project = self.project_controller.get(name=project_name)
        return db.Skills.query.filter(
            db.Skills.name == skill_name,
            db.Skills.project_id == project.id,
            db.Skills.deleted_at == null()
        ).first()

    def get_skills(self, project_name: str) -> List[dict]:
        '''
        Gets all skills in a project.

        Parameters:
            project_name (str): The name of the containing project

        Returns:
            all_skills (List[db.Skills]): List of database skill object

        Raises:
            ValueError: If `project_name` does not exist
        '''

        project_controller = ProjectController()
        projects = project_controller.get_list()
        if project_name is not None:
            projects = list([p for p in projects if p.name == project_name])
        project_ids = list([p.id for p in projects])

        query = (
            db.session.query(db.Skills)
            .filter(
                db.Skills.project_id.in_(project_ids),
                db.Skills.deleted_at == null()
            )
        )

        return query.all()

    def add_skill(
            self,
            name: str,
            project_name: str,
            type: str,
            params: Dict[str, str] = {}) -> db.Skills:
        '''
        Adds a skill to the database.

        Parameters:
            name (str): The name of the new skill
            project_name (str): The containing project
            type (str): The type of the skill (e.g. Knowledge Base)
            params: (Dict[str, str]): Parameters associated with the skill

        Returns:
            bot (db.Skills): The created skill

        Raises:
            ValueError: If `project_name` does not exist or skill already exists
        '''
        if project_name is None:
            project_name = 'mindsdb'
        project = self.project_controller.get(name=project_name)

        skill = self.get_skill(name, project_name)

        if skill is not None:
            raise ValueError(f'Skill with name already exists: {name}')

        new_skill = db.Skills(
            name=name,
            project_id=project.id,
            type=type,
            params=params,
        )
        db.session.add(new_skill)
        db.session.commit()

        return new_skill

    def update_skill(
            self,
            skill_name: str,
            new_name: str = None,
            project_name: str = 'mindsdb',
            type: str = None,
            params: Dict[str, str] = None):
        '''
        Updates an existing skill in the database.

        Parameters:
            skill_name (str): The name of the new skill, or existing skill to update
            new_name (str): Updated name of the skill
            project_name (str): The containing project
            type (str): The type of the skill (e.g. Knowledge Base)
            params: (Dict[str, str]): Parameters associated with the skill

        Returns:
            bot (db.Skills): The updated skill

        Raises:
            ValueError: If `project_name` does not exist or skill doesn't exist
        '''

        existing_skill = self.get_skill(skill_name, project_name)
        if existing_skill is None:
            raise ValueError(f'Skill with name not found: {skill_name}')
        if isinstance(existing_skill.params, dict) and existing_skill.params.get('is_demo') is True:
            raise ValueError("It is forbidden to change properties of the demo object")

        if new_name is not None:
            existing_skill.name = new_name
        if type is not None:
            existing_skill.type = type
        if params is not None:
            # Merge params on update
            existing_params = {} if not existing_skill.params else existing_skill.params
            existing_params.update(params)
            # Remove None values entirely.
            params = {k: v for k, v in existing_params.items() if v is not None}
            existing_skill.params = params
            # Some versions of SQL Alchemy won't handle JSON updates correctly without this.
            # See: https://docs.sqlalchemy.org/en/20/orm/session_api.html#sqlalchemy.orm.attributes.flag_modified
            flag_modified(existing_skill, 'params')

        db.session.commit()

        return existing_skill

    def delete_skill(self, skill_name: str, project_name: str = 'mindsdb'):
        '''
        Deletes a skill by name.

        Parameters:
            skill_name (str): The name of the skill to delete
            project_name (str): The name of the containing project

        Raises:
            ValueError: If `project_name` does not exist or skill doesn't exist
        '''

        skill = self.get_skill(skill_name, project_name)
        if skill is None:
            raise ValueError(f"Skill with name doesn't exist: {skill_name}")
        if isinstance(skill.params, dict) and skill.params.get('is_demo') is True:
            raise ValueError("Unable to delete demo object")
        skill.deleted_at = datetime.datetime.now()
        db.session.commit()



"""
This is the GreptimeDB integration handler for mindsdb.  It provides the routines
which provide for interacting with the database.

Because GreptimeDB has built-in MySQL wire protocol support, this handler is simply
 a subclass of mindsdb's MySQL handler
"""

from mindsdb.integrations.handlers.mysql_handler import Handler as MySQLHandler


class GreptimeDBHandler(MySQLHandler):
    """
    This handler handles connection and execution of GreptimeDB statements.
    It's a subclass of default mysql handler
    """

    name = 'greptimedb'

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)



import ast
from typing import List, Optional

import numpy as np
from pinecone import Pinecone, ServerlessSpec
from pinecone.core.openapi.shared.exceptions import NotFoundException, PineconeApiException
import pandas as pd

from mindsdb.integrations.libs.response import RESPONSE_TYPE
from mindsdb.integrations.libs.response import HandlerResponse
from mindsdb.integrations.libs.response import HandlerResponse as Response
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse
from mindsdb.integrations.libs.vectordatabase_handler import (
    FilterCondition,
    FilterOperator,
    TableField,
    VectorStoreHandler,
)
from mindsdb.utilities import log

logger = log.getLogger(__name__)

DEFAULT_CREATE_TABLE_PARAMS = {
    "dimension": 8,
    "metric": "cosine",
    "spec": {
        "cloud": "aws",
        "region": "us-east-1"
    }
}
MAX_FETCH_LIMIT = 10000
UPSERT_BATCH_SIZE = 99  # API reccomendation


class PineconeHandler(VectorStoreHandler):
    """This handler handles connection and execution of the Pinecone statements."""

    name = "pinecone"

    def __init__(self, name: str, connection_data: dict, **kwargs):
        super().__init__(name)
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def _get_index_handle(self, index_name):
        """Returns handler to index specified by `index_name`"""
        connection = self.connect()
        index = connection.Index(index_name)
        try:
            index.describe_index_stats()
        except Exception:
            index = None
        return index

    def _get_pinecone_operator(self, operator: FilterOperator) -> str:
        """Convert FilterOperator to an operator that pinecone's query language can undersand"""
        mapping = {
            FilterOperator.EQUAL: "$eq",
            FilterOperator.NOT_EQUAL: "$ne",
            FilterOperator.GREATER_THAN: "$gt",
            FilterOperator.GREATER_THAN_OR_EQUAL: "$gte",
            FilterOperator.LESS_THAN: "$lt",
            FilterOperator.LESS_THAN_OR_EQUAL: "$lte",
            FilterOperator.IN: "$in",
            FilterOperator.NOT_IN: "$nin",
        }
        if operator not in mapping:
            raise Exception(f"Operator {operator} is not supported by Pinecone!")
        return mapping[operator]

    def _translate_metadata_condition(self, conditions: List[FilterCondition]) -> Optional[dict]:
        """
        Translate a list of FilterCondition objects a dict that can be used by pinecone.
        E.g.,
        [
            FilterCondition(
                column="metadata.created_at",
                op=FilterOperator.LESS_THAN,
                value="2020-01-01",
            ),
            FilterCondition(
                column="metadata.created_at",
                op=FilterOperator.GREATER_THAN,
                value="2019-01-01",
            )
        ]
        -->
        {
            "$and": [
                {"created_at": {"$lt": "2020-01-01"}},
                {"created_at": {"$gt": "2019-01-01"}}
            ]
        }
        """
        # we ignore all non-metadata conditions
        if conditions is None:
            return None
        metadata_conditions = [
            condition
            for condition in conditions
            if condition.column.startswith(TableField.METADATA.value)
        ]
        if len(metadata_conditions) == 0:
            return None

        # we translate each metadata condition into a dict
        pinecone_conditions = []
        for condition in metadata_conditions:
            metadata_key = condition.column.split(".")[-1]
            pinecone_conditions.append(
                {
                    metadata_key: {
                        self._get_pinecone_operator(condition.op): condition.value
                    }
                }
            )

        # we combine all metadata conditions into a single dict
        metadata_condition = (
            {"$and": pinecone_conditions}
            if len(pinecone_conditions) > 1
            else pinecone_conditions[0]
        )
        return metadata_condition

    def _matches_to_dicts(self, matches: List):
        """Converts the custom pinecone response type to a list of python dict"""
        return [match.to_dict() for match in matches]

    def connect(self):
        """Connect to a pinecone database."""
        if self.is_connected is True:
            return self.connection

        if 'api_key' not in self.connection_data:
            raise ValueError('Required parameter (api_key) must be provided.')

        try:
            self.connection = Pinecone(api_key=self.connection_data['api_key'])
            return self.connection
        except Exception as e:
            logger.error(f"Error connecting to Pinecone client, {e}!")
            self.is_connected = False

    def disconnect(self):
        """Close the pinecone connection."""
        if self.is_connected is False:
            return
        self.connection = None
        self.is_connected = False

    def check_connection(self):
        """Check the connection to pinecone."""
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            connection = self.connect()
            connection.list_indexes()
            response.success = True
        except Exception as e:
            logger.error(f"Error connecting to pinecone , {e}!")
            response.error_message = str(e)

        if response.success is True and need_to_close:
            self.disconnect()
        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def get_tables(self) -> HandlerResponse:
        """Get the list of indexes in the pinecone database."""
        connection = self.connect()
        indexes = connection.list_indexes()
        df = pd.DataFrame(
            columns=["table_name"],
            data=[index['name'] for index in indexes],
        )
        return Response(resp_type=RESPONSE_TYPE.TABLE, data_frame=df)

    def create_table(self, table_name: str, if_not_exists=True):
        """Create an index with the given name in the Pinecone database."""
        connection = self.connect()

        # TODO: Should other parameters be supported? Pod indexes?
        # TODO: Should there be a better way to provide these parameters rather than when establishing the connection?
        create_table_params = {}
        for key, val in DEFAULT_CREATE_TABLE_PARAMS.items():
            if key in self.connection_data:
                create_table_params[key] = self.connection_data[key]
            else:
                create_table_params[key] = val

        create_table_params["spec"] = ServerlessSpec(**create_table_params["spec"])

        try:
            connection.create_index(name=table_name, **create_table_params)
        except PineconeApiException as pinecone_error:
            if pinecone_error.status == 409 and if_not_exists:
                return
            raise Exception(f"Error creating index '{table_name}': {pinecone_error}")

    def insert(self, table_name: str, data: pd.DataFrame):
        """Insert data into pinecone index passed in through `table_name` parameter."""
        index = self._get_index_handle(table_name)
        if index is None:
            raise Exception(f"Error getting index '{table_name}', are you sure the name is correct?")

        data.rename(columns={
            TableField.ID.value: "id",
            TableField.EMBEDDINGS.value: "values"},
            inplace=True)

        columns = ["id", "values"]

        if TableField.METADATA.value in data.columns:
            data.rename(columns={TableField.METADATA.value: "metadata"}, inplace=True)
            # fill None and NaN values with empty dict
            if data['metadata'].isnull().any():
                data['metadata'] = data['metadata'].apply(lambda x: {} if x is None or (isinstance(x, float) and np.isnan(x)) else x)
            columns.append("metadata")

        data = data[columns]

        # convert the embeddings to lists if they are strings
        data["values"] = data["values"].apply(lambda x: ast.literal_eval(x) if isinstance(x, str) else x)

        for chunk in (data[pos:pos + UPSERT_BATCH_SIZE] for pos in range(0, len(data), UPSERT_BATCH_SIZE)):
            chunk = chunk.to_dict(orient="records")
            index.upsert(vectors=chunk)

    def drop_table(self, table_name: str, if_exists=True):
        """Delete an index passed in through `table_name` from the pinecone ."""
        connection = self.connect()
        try:
            connection.delete_index(table_name)
        except NotFoundException:
            if if_exists:
                return
            raise Exception(f"Error deleting index '{table_name}', are you sure the name is correct?")

    def delete(self, table_name: str, conditions: List[FilterCondition] = None):
        """Delete records in pinecone index `table_name` based on ids or based on metadata conditions."""
        filters = self._translate_metadata_condition(conditions)
        ids = [
            condition.value
            for condition in conditions
            if condition.column == TableField.ID.value
        ] or None
        if filters is None and ids is None:
            raise Exception("Delete query must have either id condition or metadata condition!")
        index = self._get_index_handle(table_name)
        if index is None:
            raise Exception(f"Error getting index '{table_name}', are you sure the name is correct?")

        if filters is None:
            index.delete(ids=ids)
        else:
            index.delete(filter=filters)

    def select(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        offset: int = None,
        limit: int = None,
    ):
        """Run query on pinecone index named `table_name` and get results."""
        # TODO: Add support for namespaces.
        index = self._get_index_handle(table_name)
        if index is None:
            raise Exception(f"Error getting index '{table_name}', are you sure the name is correct?")

        query = {
            "include_values": True,
            "include_metadata": True
        }

        # check for metadata filter
        metadata_filters = self._translate_metadata_condition(conditions)
        if metadata_filters is not None:
            query["filter"] = metadata_filters

        # check for vector and id filters
        vector_filters = []
        id_filters = []

        if conditions:
            for condition in conditions:
                if condition.column == TableField.SEARCH_VECTOR.value:
                    vector_filters.append(condition.value)
                elif condition.column == TableField.ID.value:
                    id_filters.append(condition.value)

        if vector_filters:
            if len(vector_filters) > 1:
                raise Exception("You cannot have multiple search_vectors in query")

            query["vector"] = vector_filters[0]
            # For subqueries, the vector filter is a list of list of strings
            if isinstance(query["vector"], list) and isinstance(query["vector"][0], str):
                if len(query["vector"]) > 1:
                    raise Exception("You cannot have multiple search_vectors in query")

                try:
                    query["vector"] = ast.literal_eval(query["vector"][0])
                except Exception as e:
                    raise Exception(f"Cannot parse the search vector '{query['vector']}'into a list: {e}")

        if id_filters:
            if len(id_filters) > 1:
                raise Exception("You cannot have multiple IDs in query")

            query["id"] = id_filters[0]

        if not vector_filters and not id_filters:
            raise Exception("You must provide either a search_vector or an ID in the query")

        # check for limit
        if limit is not None:
            query["top_k"] = limit
        else:
            query["top_k"] = MAX_FETCH_LIMIT

        # exec query
        try:
            result = index.query(**query)
        except Exception as e:
            raise Exception(f"Error running SELECT query on '{table_name}': {e}")

        # convert to dataframe
        df_columns = {
            "id": TableField.ID.value,
            "metadata": TableField.METADATA.value,
            "values": TableField.EMBEDDINGS.value,
        }
        results_df = pd.DataFrame.from_records(self._matches_to_dicts(result["matches"]))
        if bool(len(results_df.columns)):
            results_df.rename(columns=df_columns, inplace=True)
        else:
            results_df = pd.DataFrame(columns=list(df_columns.values()))
        results_df[TableField.CONTENT.value] = ""
        return results_df[columns]

    def get_columns(self, table_name: str) -> HandlerResponse:
        return super().get_columns(table_name)



from __future__ import annotations

import logging
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    Optional,
)

import pandas as pd
from langchain_core.callbacks import (
    CallbackManagerForLLMRun,
)
from langchain_core.language_models.chat_models import (
    BaseChatModel,
)
from langchain_core.messages import (
    AIMessage,
    BaseMessage,
    ChatMessage,
    FunctionMessage,
    HumanMessage,
    SystemMessage,
)
from langchain_core.outputs import (
    ChatGeneration,
    ChatResult,
)
from pydantic import model_validator

from mindsdb.interfaces.agents.constants import USER_COLUMN

logger = logging.getLogger(__name__)


def _convert_message_to_dict(message: BaseMessage) -> dict:
    if isinstance(message, ChatMessage):
        message_dict = {"role": message.role, "content": message.content}
    elif isinstance(message, HumanMessage):
        message_dict = {"role": "user", "content": message.content}
    elif isinstance(message, AIMessage):
        message_dict = {"role": "assistant", "content": message.content}
        if "function_call" in message.additional_kwargs:
            message_dict["function_call"] = message.additional_kwargs["function_call"]
    elif isinstance(message, SystemMessage):
        message_dict = {"role": "system", "content": message.content}
    elif isinstance(message, FunctionMessage):
        message_dict = {
            "role": "function",
            "content": message.content,
            "name": message.name,
        }
    else:
        raise ValueError(f"Got unknown type {message}")
    if "name" in message.additional_kwargs:
        message_dict["name"] = message.additional_kwargs["name"]
    return message_dict


class ChatMindsdb(BaseChatModel):
    """A chat model that uses the Mindsdb"""

    model_name: str
    project_name: Optional[str] = 'mindsdb'
    model_info: Optional[dict] = None
    project_datanode: Optional[Any] = None

    class Config:
        """Configuration for this pydantic object."""
        arbitrary_types_allowed = True
        allow_reuse = True

    @property
    def _default_params(self) -> Dict[str, Any]:
        return {}

    def completion(
            self, messages: List[dict]
    ) -> Any:
        problem_definition = self.model_info['problem_definition'].get('using', {})
        output_col = self.model_info['predict']

        # TODO create table for conversational model?
        if len(messages) > 1:
            content = '\n'.join([
                f"{m['role']}: {m['content']}"
                for m in messages
            ])
        else:
            content = messages[0]['content']

        record = {}
        params = {}
        # Default to conversational if not set.
        mode = problem_definition.get('mode', 'conversational')
        if mode == 'conversational' or mode == 'retrieval':
            # flag for langchain to prevent calling agent inside of agent
            if self.model_info['engine'] == 'langchain':
                params['mode'] = 'chat_model'

            user_column = problem_definition.get('user_column', USER_COLUMN)
            record[user_column] = content

        elif 'column' in problem_definition:
            # input defined as 'column' param
            record[problem_definition['column']] = content

        else:
            # failback, maybe handler supports template injection
            params['prompt_template'] = content

        predictions = self.project_datanode.predict(
            model_name=self.model_name,
            df=pd.DataFrame([record]),
            params=params,
        )

        col = output_col
        if col not in predictions.columns:
            # get first column
            col = predictions.columns[0]

        # get first row
        result = predictions[col][0]

        # TODO token calculation
        return {
            'messages': [result]
        }

    @model_validator(mode='before')
    def validate_environment(cls, values: Dict) -> Dict:

        model_name = values['model_name']
        project_name = values['project_name']

        from mindsdb.api.executor.controllers import SessionController

        session = SessionController()
        session.database = 'mindsdb'

        values['model_info'] = session.model_controller.get_model(model_name, project_name=project_name)

        project_datanode = session.datahub.get(values['project_name'])

        values["project_datanode"] = project_datanode

        return values

    def _generate(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            run_manager: Optional[CallbackManagerForLLMRun] = None,
            stream: Optional[bool] = None,
            **kwargs: Any,
    ) -> ChatResult:

        message_dicts = [_convert_message_to_dict(m) for m in messages]

        response = self.completion(
            messages=message_dicts
        )
        return self._create_chat_result(response)

    def _create_chat_result(self, response: Mapping[str, Any]) -> ChatResult:
        generations = []
        for content in response["messages"]:
            message = AIMessage(content=content)
            gen = ChatGeneration(
                message=message,
                generation_info=dict(finish_reason=None),
            )
            generations.append(gen)
        token_usage = response.get("usage", {})
        set_model_value = self.model_name
        if self.model_name is not None:
            set_model_value = self.model_name
        llm_output = {"token_usage": token_usage, "model": set_model_value}
        return ChatResult(generations=generations, llm_output=llm_output)

    @property
    def _identifying_params(self) -> Dict[str, Any]:
        """Get the identifying parameters."""
        set_model_value = self.model_name
        if self.model_name is not None:
            set_model_value = self.model_name
        return {
            "model_name": set_model_value,
        }

    @property
    def _llm_type(self) -> str:
        return "mindsdb"



class InvalidOpenAIModel(Exception):
    pass


class InvalidWriterModel(Exception):
    pass


class MissingPromptTemplate(Exception):
    pass


class UnsupportedVectorStore(Exception):
    pass


class MissingUseIndex(Exception):
    pass


class UnsupportedLLM(Exception):
    pass


class InvalidPromptTemplate(Exception):
    pass



import unittest
from mindsdb.api.executor.data_types.response_type import (
    RESPONSE_TYPE,
)
from mindsdb.integrations.handlers.cloud_spanner_handler.cloud_spanner_handler import (
    CloudSpannerHandler,
)


class CloudSpannerHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {'connection_data': {'database_id': 'example-db', 'instance_id': 'test-instance', 'project': 'your-project-id'}}
        cls.handler = CloudSpannerHandler('test_cloud_spanner_handler', **cls.kwargs)


    def test_0_connect(self):
        self.handler.connect()

    def test_1_check_connection(self):
        self.handler.check_connection()

    def test_2_create_table(self):
        res = self.handler.query('CREATE TABLE integers(i INT64) PRIMARY KEY (i)')
        assert res.type is not RESPONSE_TYPE.ERROR


    def test_3_insert_into_table(self):
        res = self.handler.query('INSERT INTO integers (i) VALUES (42)')
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_4_select(self):
        res = self.handler.query('SELECT * FROM integers')
        assert res.type is RESPONSE_TYPE.TABLE

    def test_5_describe_table(self):
        res = self.handler.get_columns('integers')
        assert res.type is RESPONSE_TYPE.TABLE

    def test_6_drop_table(self):
        res = self.handler.query('DROP TABLE integers')
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_7_get_tables(self):
        res = self.handler.get_tables()
        assert res.type is not RESPONSE_TYPE.ERROR


if __name__ == '__main__':
    unittest.main()



from enum import Enum

from typing import List, Dict, Any, Optional, Union, Callable


from pydantic import BaseModel, Field, model_validator


from mindsdb.integrations.utilities.rag.settings import DEFAULT_CHUNK_OVERLAP, DEFAULT_CHUNK_SIZE
from mindsdb.integrations.utilities.rag.settings import LLMConfig


class PreprocessorType(Enum):
    CONTEXTUAL = "contextual"
    TEXT_CHUNKING = "text_chunking"


class BasePreprocessingConfig(BaseModel):
    """Base configuration for preprocessing"""
    chunk_size: int = Field(default=DEFAULT_CHUNK_SIZE, description="Size of document chunks")
    chunk_overlap: int = Field(default=DEFAULT_CHUNK_OVERLAP, description="Overlap between chunks")


class ContextualConfig(BasePreprocessingConfig):
    """Configuration specific to contextual preprocessing"""
    llm_config: LLMConfig = Field(
        default_factory=LLMConfig,
        description="LLM configuration to use for context generation"
    )
    context_template: Optional[str] = Field(
        default=None,
        description="Custom template for context generation"
    )
    summarize: Optional[bool] = Field(
        default=False,
        description="Whether to return chunks as summarizations"
    )


class TextChunkingConfig(BaseModel):
    """Configuration for text chunking preprocessor using Pydantic"""
    chunk_size: int = Field(
        default=1000,
        description="The target size of each text chunk",
        gt=0
    )
    chunk_overlap: int = Field(
        default=200,
        description="The number of characters to overlap between chunks",
        ge=0
    )
    length_function: Callable = Field(
        default=len,
        description="Function to measure text length"
    )
    separators: List[str] = Field(
        default=["\n\n", "\n", " ", ""],
        description="List of separators to use for splitting text, in order of priority"
    )

    class Config:
        arbitrary_types_allowed = True


class PreprocessingConfig(BaseModel):
    """Complete preprocessing configuration"""
    type: PreprocessorType = Field(
        default=PreprocessorType.TEXT_CHUNKING,
        description="Type of preprocessing to apply"
    )
    contextual_config: Optional[ContextualConfig] = Field(
        default=None,
        description="Configuration for contextual preprocessing"
    )
    text_chunking_config: Optional[TextChunkingConfig] = Field(
        default=None,
        description="Configuration for text chunking preprocessing"
    )

    @model_validator(mode='after')
    def validate_config_presence(self) -> 'PreprocessingConfig':
        """Ensure the appropriate config is present for the chosen type"""
        if self.type == PreprocessorType.CONTEXTUAL and not self.contextual_config:
            self.contextual_config = ContextualConfig()
        if self.type == PreprocessorType.TEXT_CHUNKING and not self.text_chunking_config:
            self.text_chunking_config = TextChunkingConfig()
        return self


class Document(BaseModel):

    """Document model with default metadata handling"""
    id: Optional[Union[int, str]] = Field(default=None, description="Unique identifier for the document")
    content: str = Field(description="The document content")
    embeddings: Optional[List[float]] = Field(default=None, description="Vector embeddings of the content")
    metadata: Optional[Dict[str, Any]] = Field(default=None, description="Additional document metadata")

    @model_validator(mode='after')
    def validate_metadata(self) -> 'Document':
        """Ensure metadata is present and valid"""
        if not self.metadata:
            self.metadata = {'source': 'default'}
        return self


class ProcessedChunk(Document):
    """Processed chunk that aligns with VectorStoreHandler schema"""
    pass



from mindsdb_sql_parser import parse_sql
import pandas as pd
from mendeley import Mendeley
from mindsdb.integrations.libs.api_handler import APIHandler, FuncParser
from mendeley.session import MendeleySession
from mindsdb.integrations.handlers.mendeley_handler.mendeley_tables import CatalogSearchTable 
from mindsdb.utilities import log
from typing import Dict
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)

logger = log.getLogger(__name__)

class MendeleyHandler(APIHandler): 

    def __init__(self, name, **kwargs): 

        """ constructor
        Args:
            name (str): the handler name
        """
        super().__init__(name)

        self.connection_args = kwargs.get('connection_data', {})
        
        self.client_id = self.connection_args.get('client_id', None)
        self.client_secret = self.connection_args.get('client_secret', None)
        self.session = self.connect()

        self.session = None
        self.is_connected = False
    
        catalog_search_data = CatalogSearchTable(self)
        self.catalog_search_data = catalog_search_data 
        self._register_table('catalog_search_data', catalog_search_data)

    def connect(self) -> MendeleySession:
        """ The connect method sets up the connection required by the handler.
        In order establish a connection with Mendeley API one needs the client id and client secret that are
        created after registering the application at https://dev.mendeley.com/myapps.html . More information on the matter
        can be found at https://dev.mendeley.com/reference/topics/application_registration.html .
        In order to have access to Mendeley data we use "session".

        Returns:
        HandlerStatusResponse """

        if self.is_connected == True:
            return self.session

        mendeley = Mendeley(self.client_id, self.client_secret)
        auth = mendeley.start_client_credentials_flow()
        self.session = auth.authenticate()

        self.is_connected = True
        return self.session
    
    def check_connection(self) -> StatusResponse:

        """ The check_connection method checks the connection to the handler
        Returns:
            HandlerStatusResponse
        """
        response = StatusResponse(False)

        try:
            self.connect()
            response.success = True

        except Exception as e:
            logger.error(f'Error connecting to Mendeley: {e}!')
            response.error_message = str(e)

        self.is_connected = response.success
        return response

    def native_query(self, query_string: str):
        
        """The native_query method receives raw query and acts upon it.
            Args:
                query_string (str): query in native format
            Returns:
                HandlerResponse
        """
        ast = parse_sql(query_string)
        return self.query(ast)
    
    def get_authors(self,data):
        """The get_authors method receives the data - a specific document returned by the API, gets the names of the authors 
            and combines them in a string, so as to allow the use of DataFrame.
            Args:
                data (CatalogDocument): document returned by API
            Returns:
                authors string
        """
        authors = ""
        sum = 0
        if data.authors!=None:
            for x in data.authors:
                if sum + 1 == len(data.authors) and x.first_name!=None and x.last_name!=None  :
                    authors = authors + x.first_name + " " + x.last_name
                else:
                    if x.first_name!=None and x.last_name!=None:
                        authors = authors + x.first_name + " " + x.last_name + ", "
                        sum = sum + 1
        return authors

    def get_keywords(self,data):
        """The get_keywords method receives the data-a specific document returned by the API, gets the specified keywords 
            and combines them in a string, so as to allow the use of DataFrame.
            Args:
                data (CatalogDocument) : document returned by the API
            Returns:
                keywords string
        """
        keywords = ""
        sum = 0
        if data.keywords!=None:
            for x in data.keywords:
                if sum + 1 == len(data.keywords):
                    keywords = keywords  + x + " "
                else:
                    if x!=None :
                        keywords = keywords + x + ", "
                        sum = sum + 1
        return keywords
    
    def create_dict(self,data):

        """The create_dict method receives the data-a specific document returned by the API, gets the resources-fields of the document,
          as specified in Mendley documentation, and puts them in a dictionary.

            Args:
                data (CatalogDocument) : document returned by API
            Returns:
                dict dictionary
        """
        dict = {}
        dict["title"] = data.title
        dict["type"] = data.type
        dict["source"] = data.source
        dict["year"] = data.year
        if data.identifiers!=None:
            dict["pmid"] = data.identifiers.get("pmid")
            dict["sgr"] = data.identifiers.get("sgr")
            dict["issn"] = data.identifiers.get("issn")
            dict["scopus"] = data.identifiers.get("scopus")
            dict["doi"] = data.identifiers.get("doi")
            dict["pui"] = data.identifiers.get("pui")
        dict["authors"] = self.get_authors(data)
        if data.keywords!=None:
            dict["keywords"] = self.get_keywords(data)
        else:
            dict["keywords"] = None
        dict["link"] = data.link
        dict["id"] = data.id
        return dict
    
    def call_mendeley_api(self, method_name: str, params: Dict) -> pd.DataFrame:
        
        """The method call_mendeley_api is used to communicate with Mendeley. Depending on the method used there are three different types
        of search conducted.
        The advanced_search results in a CatalogSearch resource, which, depending on the parameters used, could either be a number of different documents (CatalogDocument),
        a single one or none.
        The by_identifier search is more specific in nature and can result either in one or no CatalogDocuments.
        The get search has the same results as the by_identifier.
        If the method specified does not exist, an NotImplementedError is raised.
        Args:
            method_name (str) : name of method
            params (Dict): Dictionary containing the parameters used in the search
        Returns:
            DataFrame
        """

        self.session= self.connect()

        if method_name == 'advanced_search':
            search_params = {
            'title': params.get("title"),
            'author': params.get("author"),
            'source': params.get("source"),
            'abstract': params.get("abstract"),
            'min_year': params.get("min_year"),
            'max_year': params.get("max_year"),
            'open_access': params.get("open_access")
            }
            data = self.session.catalog.advanced_search(**search_params)
            sum = 0
            df = pd.DataFrame()
            for x in data.list(page_size=params["limit"]).items:
                    if sum == 0:
                        df = pd.DataFrame(self.create_dict(x),index=[0])
                        sum += 1
                    else : 
                        df = df.append(self.create_dict(x),ignore_index = True)
                        sum += 1
            if df.empty:
                raise NotImplementedError(('Insufficient or wrong input given'))
            else:
                return df

        elif method_name == 'identifier_search':
            search_params = {
            'arxiv': params.get("arxiv"),
            'doi': params.get("doi"),
            'isbn': params.get("isbn"),
            'issn': params.get("issn"),
            'pmid': params.get("pmid"),
            'scopus': params.get("scopus"),
            'filehash': params.get("filehash")
            }
            data = self.session.catalog.by_identifier(**search_params)
            df = pd.DataFrame(self.create_dict(data),index=[0])
            return df

        elif method_name == 'get':
            data = self.session.catalog.get(params.get("id"))
            df = pd.DataFrame(self.create_dict(data),index=[0])
            return df

        raise NotImplementedError('Method name {} not supported by Mendeley API Handler'.format(method_name))





import concurrent.futures
import io
import re
import traceback
from threading import Lock
from typing import List
from urllib.parse import urljoin, urlparse, urlunparse

import html2text
import fitz  # PyMuPDF
import pandas as pd
import requests
from bs4 import BeautifulSoup
from mindsdb.utilities import log

logger = log.getLogger(__name__)


def pdf_to_markdown(response, gap_threshold=10):
    """
    Convert a PDF document to Markdown text.

    Args:
        response: the response object containing the PDF data
        gap_threshold (int): the vertical gap size that triggers a new line in the output (default 10)

    Returns:
        A string containing the converted Markdown text.

    Raises:
        Exception -- if the PDF data cannot be processed.
    """

    try:
        file_stream = io.BytesIO(response.content)
        document = fitz.open(stream=file_stream, filetype="pdf")
    except Exception as e:
        raise Exception("Failed to process PDF data: " + str(e))

    markdown_lines = []
    for page_num in range(len(document)):
        page = document.load_page(page_num)

        blocks = page.get_text("blocks")

        blocks.sort(key=lambda block: (block[1], block[0]))

        previous_block_bottom = 0
        for block in blocks:
            y0 = block[1]
            y1 = block[3]
            block_text = block[4]

            # Check if there's a large vertical gap between this block and the previous one
            if y0 - previous_block_bottom > gap_threshold:
                markdown_lines.append("")

            markdown_lines.append(block_text)
            previous_block_bottom = y1

        markdown_lines.append("")

    document.close()

    return "\n".join(markdown_lines)


def is_valid(url) -> bool:
    """
    Check if a URL is valid.

    Args:
        url: the URL to check

    Returns:
        bool: True if the URL is valid, False otherwise.
    """
    parsed = urlparse(url)
    return bool(parsed.netloc) and bool(parsed.scheme)


def parallel_get_all_website_links(urls) -> dict:
    """
    Fetch all website links from a list of URLs.

    Args:
        urls (list): a list of URLs to fetch links from

    Returns:
        A dictionary mapping each URL to a list of links found on that URL.

    Raises:
        Exception: if an error occurs while fetching links from a URL.
    """
    url_contents = {}

    if len(urls) <= 10:
        for url in urls:
            url_contents[url] = get_all_website_links(url)
        return url_contents

    with concurrent.futures.ProcessPoolExecutor() as executor:
        future_to_url = {
            executor.submit(get_all_website_links, url): url for url in urls
        }
        for future in concurrent.futures.as_completed(future_to_url):
            url = future_to_url[future]
            try:
                url_contents[url] = future.result()
            except Exception as exc:
                logger.error(f'{url} generated an exception: {exc}')
                # don't raise the exception, just log it, continue processing other urls

    return url_contents


def get_all_website_links(url) -> dict:
    """
    Fetch all website links from a URL.

    Args:
        url (str): the URL to fetch links from

    Returns:
        A dictionary containing the URL, the extracted links, the HTML content, the text content, and any error that occurred.
    """
    logger.info("rawling: {url} ...".format(url=url))
    urls = set()

    domain_name = urlparse(url).netloc
    try:
        session = requests.Session()

        # Add headers to mimic a real browser request
        headers = {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/74.0.3729.169 Safari/537.36"
        }

        response = session.get(url, headers=headers)
        if "cookie" in response.request.headers:
            session.cookies.update(response.cookies)

        content_type = response.headers.get("Content-Type", "").lower()

        if "application/pdf" in content_type:
            content_html = "PDF"
            content_text = pdf_to_markdown(response)
        else:
            content_html = response.text

            # Parse HTML content with BeautifulSoup
            soup = BeautifulSoup(content_html, "html.parser")
            content_text = get_readable_text_from_soup(soup)
            for a_tag in soup.findAll("a"):
                href = a_tag.attrs.get("href")
                if href == "" or href is None:
                    continue
                href = urljoin(url, href)
                parsed_href = urlparse(href)
                href = urlunparse((parsed_href.scheme, parsed_href.netloc, parsed_href.path, '', '', ''))
                if not is_valid(href):
                    continue
                if href in urls:
                    continue
                if domain_name != parsed_href.netloc:
                    continue

                href = href.rstrip("/")
                urls.add(href)

    except Exception as e:
        error_message = traceback.format_exc().splitlines()[-1]
        logger.error("An exception occurred: %s", str(e))
        return {
            "url": url,
            "urls": urls,
            "html_content": "",
            "text_content": "",
            "error": str(error_message),
        }

    return {
        "url": url,
        "urls": urls,
        "html_content": content_html,
        "text_content": content_text,
        "error": None,
    }


def get_readable_text_from_soup(soup) -> str:
    """
    Extract readable text from a BeautifulSoup object and convert it to Markdown.

    Args:
        soup (BeautifulSoup): a BeautifulSoup object

    Returns:
        The extracted text in Markdown format.
    """
    html_converter = html2text.HTML2Text()
    html_converter.ignore_links = False
    return html_converter.handle(str(soup))


def get_all_website_links_recursively(url, reviewed_urls, limit=None, crawl_depth: int = 1, current_depth: int = 0, filters: List[str] = None):
    """
    Recursively gathers all links from a given website up to a specified limit.

    Args:
        url (str): The starting URL to fetch links from.
        reviewed_urls (dict): A dictionary to keep track of reviewed URLs and associated data.
        limit (int, optional): The maximum number of URLs to process.
        crawl_depth: How deep to crawl from each base URL. 0 = scrape given URLs only
        current_depth: How deep we are currently crawling from the base URL.
        filters (List[str]): Crawl URLs that only match these regex patterns.

    TODO: Refactor this function to use a iterative aproach instead of recursion
    """
    if limit is not None:
        if len(reviewed_urls) >= limit:
            return reviewed_urls
    if crawl_depth == current_depth:
        return reviewed_urls

    if not filters:
        matches_filter = True
    else:
        matches_filter = any(re.match(f, url) is not None for f in filters)
    if url not in reviewed_urls and matches_filter:
        try:
            reviewed_urls[url] = get_all_website_links(url)
        except Exception as e:
            error_message = traceback.format_exc().splitlines()[-1]
            logger.error("An exception occurred: %s", str(e))
            reviewed_urls[url] = {
                "url": url,
                "urls": [],
                "html_content": "",
                "text_content": "",
                "error": str(error_message),
            }

    to_rev_url_list = []

    # create a list of new urls to review that don't exist in the already reviewed ones
    for new_url in reviewed_urls[url]["urls"]:
        if not filters:
            matches_filter = True
        else:
            matches_filter = any(re.match(f, new_url) is not None for f in filters)
        if not matches_filter:
            continue
        # if this is already in the urls, then no need to go and crawl for it
        if new_url in reviewed_urls or new_url in to_rev_url_list:
            continue

        # insert immediately to count limit between threads. fill later
        url_list_lock = Lock()
        with url_list_lock:
            if limit is None or len(reviewed_urls) < limit:
                reviewed_urls[new_url] = {}
                to_rev_url_list.append(new_url)
            else:
                break

    if len(to_rev_url_list) > 0:
        new_revised_urls = parallel_get_all_website_links(to_rev_url_list)

        reviewed_urls.update(new_revised_urls)

        for new_url in new_revised_urls:
            get_all_website_links_recursively(new_url, reviewed_urls, limit, crawl_depth=crawl_depth, current_depth=current_depth + 1, filters=filters)


def get_all_websites(urls, limit=1, html=False, crawl_depth: int = 1, filters: List[str] = None) -> pd.DataFrame:
    """
    Crawl a list of websites and return a DataFrame containing the results.

    Args:
        urls (list): a list of URLs to crawl
        limit (int): Absolute max number of web pages to crawl, regardless of crawl depth.
        crawl_depth (int): Crawl depth for URLs.
        html (bool): a boolean indicating whether to include the HTML content in the results
        filters (List[str]): Crawl URLs that only match these regex patterns.

    Returns:
        A DataFrame containing the results.
    """
    reviewed_urls = {}

    def fetch_url(url, crawl_depth: int = 1, filters: List[str] = None):
        # Allow URLs to be passed wrapped in quotation marks so they can be used
        # directly from the SQL editor.
        if url.startswith("'") and url.endswith("'"):
            url = url[1:-1]
        url = url.rstrip("/")
        if urlparse(url).scheme == "":
            # Try HTTPS first
            url = "https://" + url
        get_all_website_links_recursively(url, reviewed_urls, limit, crawl_depth=crawl_depth, filters=filters)

    # Use a ThreadPoolExecutor to run the helper function in parallel.
    with concurrent.futures.ThreadPoolExecutor() as executor:
        future_to_url = {executor.submit(fetch_url, url, crawl_depth=crawl_depth, filters=filters): url for url in urls}

        for future in concurrent.futures.as_completed(future_to_url):
            future.result()

    columns_to_ignore = ["urls"]
    if html is False:
        columns_to_ignore += ["html_content"]
    df = dict_to_dataframe(
        reviewed_urls, columns_to_ignore=columns_to_ignore, index_name="url"
    )

    if not df.empty and df[df.error.isna()].empty:
        raise Exception(str(df.iloc[0].error))
    return df


def dict_to_dataframe(dict_of_dicts, columns_to_ignore=None, index_name=None) -> pd.DataFrame:
    """
    Convert a dictionary of dictionaries to a DataFrame.

    Args:
        dict_of_dicts (dict): a dictionary of dictionaries
        columns_to_ignore (list): a list of columns to ignore
        index_name (str): the name of the index column
    Returns:
        A DataFrame containing the data.
    """
    df = pd.DataFrame.from_dict(dict_of_dicts, orient="index")

    if columns_to_ignore:
        for column in columns_to_ignore:
            if column in df.columns:
                df = df.drop(column, axis=1)

    if index_name:
        df.index.name = index_name

    return df



from typing import List, Dict, Text, Any

import pandas as pd
from hubspot.crm.objects import (
    SimplePublicObjectId as HubSpotObjectId,
    SimplePublicObjectBatchInput as HubSpotObjectBatchInput,
    SimplePublicObjectInputForCreate as HubSpotObjectInputCreate,
    BatchInputSimplePublicObjectId as HubSpotBatchObjectIdInput,
    BatchInputSimplePublicObjectBatchInput as HubSpotBatchObjectBatchInput,
    BatchInputSimplePublicObjectInputForCreate as HubSpotBatchObjectInputCreate,

)
from mindsdb_sql_parser import ast

from mindsdb.integrations.utilities.handlers.query_utilities import (
    INSERTQueryParser,
    SELECTQueryParser,
    UPDATEQueryParser,
    DELETEQueryParser,
    SELECTQueryExecutor,
    UPDATEQueryExecutor,
    DELETEQueryExecutor,
)

from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class CompaniesTable(APITable):
    """Hubspot Companies table."""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls Hubspot Companies data

        Parameters
        ----------
        query : ast.Select
            Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Hubspot Companies matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition

        """

        select_statement_parser = SELECTQueryParser(
            query,
            "companies",
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        companies_df = pd.json_normalize(self.get_companies(limit=result_limit))
        select_statement_executor = SELECTQueryExecutor(
            companies_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        companies_df = select_statement_executor.execute_query()

        return companies_df

    def insert(self, query: ast.Insert) -> None:
        """
        Inserts data into HubSpot "POST /crm/v3/objects/companies/batch/create" API endpoint.

        Parameters
        ----------
        query : ast.Insert
           Given SQL INSERT query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        insert_statement_parser = INSERTQueryParser(
            query,
            supported_columns=['name', 'city', 'phone', 'state', 'domain', 'industry'],
            mandatory_columns=['name'],
            all_mandatory=False,
        )
        company_data = insert_statement_parser.parse_query()
        self.create_companies(company_data)

    def update(self, query: ast.Update) -> None:
        """
        Updates data from HubSpot "PATCH /crm/v3/objects/companies/batch/update" API endpoint.

        Parameters
        ----------
        query : ast.Update
           Given SQL UPDATE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        update_statement_parser = UPDATEQueryParser(query)
        values_to_update, where_conditions = update_statement_parser.parse_query()

        companies_df = pd.json_normalize(self.get_companies())
        update_query_executor = UPDATEQueryExecutor(
            companies_df,
            where_conditions
        )

        companies_df = update_query_executor.execute_query()
        company_ids = companies_df['id'].tolist()
        self.update_companies(company_ids, values_to_update)

    def delete(self, query: ast.Delete) -> None:
        """
        Deletes data from HubSpot "DELETE /crm/v3/objects/companies/batch/archive" API endpoint.

        Parameters
        ----------
        query : ast.Delete
           Given SQL DELETE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        delete_statement_parser = DELETEQueryParser(query)
        where_conditions = delete_statement_parser.parse_query()

        companies_df = pd.json_normalize(self.get_companies())
        delete_query_executor = DELETEQueryExecutor(
            companies_df,
            where_conditions
        )

        companies_df = delete_query_executor.execute_query()
        company_ids = companies_df['id'].tolist()
        self.delete_companies(company_ids)

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_companies(limit=1)).columns.tolist()

    def get_companies(self, **kwargs) -> List[Dict]:
        hubspot = self.handler.connect()
        companies = hubspot.crm.companies.get_all(**kwargs)
        companies_dict = [
            {
                "id": company.id,
                "name": company.properties.get("name", None),
                "city": company.properties.get("city", None),
                "phone": company.properties.get("phone", None),
                "state": company.properties.get("state", None),
                "domain": company.properties.get("company", None),
                "industry": company.properties.get("industry", None),
                "createdate": company.properties["createdate"],
                "lastmodifieddate": company.properties["hs_lastmodifieddate"],
            }
            for company in companies
        ]
        return companies_dict

    def create_companies(self, companies_data: List[Dict[Text, Any]]) -> None:
        hubspot = self.handler.connect()
        companies_to_create = [HubSpotObjectInputCreate(properties=company) for company in companies_data]
        try:
            created_companies = hubspot.crm.companies.batch_api.create(
                HubSpotBatchObjectInputCreate(inputs=companies_to_create),
            )
            logger.info(f"Companies created with ID's {[created_company.id for created_company in created_companies.results]}")
        except Exception as e:
            raise Exception(f"Companies creation failed {e}")

    def update_companies(self, company_ids: List[Text], values_to_update: Dict[Text, Any]) -> None:
        hubspot = self.handler.connect()
        companies_to_update = [HubSpotObjectBatchInput(id=company_id, properties=values_to_update) for company_id in company_ids]
        try:
            updated_companies = hubspot.crm.companies.batch_api.update(
                HubSpotBatchObjectBatchInput(inputs=companies_to_update),
            )
            logger.info(f"Companies with ID {[updated_company.id for updated_company in updated_companies.results]} updated")
        except Exception as e:
            raise Exception(f"Companies update failed {e}")

    def delete_companies(self, company_ids: List[Text]) -> None:
        hubspot = self.handler.connect()
        companies_to_delete = [HubSpotObjectId(id=company_id) for company_id in company_ids]
        try:
            hubspot.crm.companies.batch_api.archive(
                HubSpotBatchObjectIdInput(inputs=companies_to_delete),
            )
            logger.info("Companies deleted")
        except Exception as e:
            raise Exception(f"Companies deletion failed {e}")


class ContactsTable(APITable):
    """Hubspot Contacts table."""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls Hubspot Contacts data

        Parameters
        ----------
        query : ast.Select
            Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Hubspot Contacts matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition

        """

        select_statement_parser = SELECTQueryParser(
            query,
            "contacts",
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        contacts_df = pd.json_normalize(self.get_contacts(limit=result_limit))
        select_statement_executor = SELECTQueryExecutor(
            contacts_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        contacts_df = select_statement_executor.execute_query()

        return contacts_df

    def insert(self, query: ast.Insert) -> None:
        """
        Inserts data into HubSpot "POST /crm/v3/objects/contacts/batch/create" API endpoint.

        Parameters
        ----------
        query : ast.Insert
           Given SQL INSERT query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        insert_statement_parser = INSERTQueryParser(
            query,
            supported_columns=['email', 'firstname', 'firstname', 'phone', 'company', 'website'],
            mandatory_columns=['email'],
            all_mandatory=False,
        )
        contact_data = insert_statement_parser.parse_query()
        self.create_contacts(contact_data)

    def update(self, query: ast.Update) -> None:
        """
        Updates data from HubSpot "PATCH /crm/v3/objects/contacts/batch/update" API endpoint.

        Parameters
        ----------
        query : ast.Update
           Given SQL UPDATE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        update_statement_parser = UPDATEQueryParser(query)
        values_to_update, where_conditions = update_statement_parser.parse_query()

        contacts_df = pd.json_normalize(self.get_contacts())
        update_query_executor = UPDATEQueryExecutor(
            contacts_df,
            where_conditions
        )

        contacts_df = update_query_executor.execute_query()
        contact_ids = contacts_df['id'].tolist()
        self.update_contacts(contact_ids, values_to_update)

    def delete(self, query: ast.Delete) -> None:
        """
        Deletes data from HubSpot "DELETE /crm/v3/objects/contacts/batch/archive" API endpoint.

        Parameters
        ----------
        query : ast.Delete
           Given SQL DELETE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        delete_statement_parser = DELETEQueryParser(query)
        where_conditions = delete_statement_parser.parse_query()

        contacts_df = pd.json_normalize(self.get_contacts())
        delete_query_executor = DELETEQueryExecutor(
            contacts_df,
            where_conditions
        )

        contacts_df = delete_query_executor.execute_query()
        contact_ids = contacts_df['id'].tolist()
        self.delete_contacts(contact_ids)

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_contacts(limit=1)).columns.tolist()

    def get_contacts(self, **kwargs) -> List[Dict]:
        hubspot = self.handler.connect()
        contacts = hubspot.crm.contacts.get_all(**kwargs)
        contacts_dict = [
            {
                "id": contact.id,
                "email": contact.properties["email"],
                "firstname": contact.properties.get("firstname", None),
                "lastname": contact.properties.get("lastname", None),
                "phone": contact.properties.get("phone", None),
                "company": contact.properties.get("company", None),
                "website": contact.properties.get("website", None),
                "createdate": contact.properties["createdate"],
                "lastmodifieddate": contact.properties["lastmodifieddate"],
            }
            for contact in contacts
        ]
        return contacts_dict

    def create_contacts(self, contacts_data: List[Dict[Text, Any]]) -> None:
        hubspot = self.handler.connect()
        contacts_to_create = [HubSpotObjectInputCreate(properties=contact) for contact in contacts_data]
        try:
            created_contacts = hubspot.crm.contacts.batch_api.create(
                HubSpotBatchObjectInputCreate(inputs=contacts_to_create)
            )
            logger.info(f"Contacts created with ID {[created_contact.id for created_contact in created_contacts.results]}")
        except Exception as e:
            raise Exception(f"Contacts creation failed {e}")

    def update_contacts(self, contact_ids: List[Text], values_to_update: Dict[Text, Any]) -> None:
        hubspot = self.handler.connect()
        contacts_to_update = [HubSpotObjectBatchInput(id=contact_id, properties=values_to_update) for contact_id in contact_ids]
        try:
            updated_contacts = hubspot.crm.contacts.batch_api.update(
                HubSpotBatchObjectBatchInput(inputs=contacts_to_update),
            )
            logger.info(f"Contacts with ID {[updated_contact.id for updated_contact in updated_contacts.results]} updated")
        except Exception as e:
            raise Exception(f"Contacts update failed {e}")

    def delete_contacts(self, contact_ids: List[Text]) -> None:
        hubspot = self.handler.connect()
        contacts_to_delete = [HubSpotObjectId(id=contact_id) for contact_id in contact_ids]
        try:
            hubspot.crm.contacts.batch_api.archive(
                HubSpotBatchObjectIdInput(inputs=contacts_to_delete),
            )
            logger.info("Contacts deleted")
        except Exception as e:
            raise Exception(f"Contacts deletion failed {e}")


class DealsTable(APITable):
    """Hubspot Deals table."""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls Hubspot Deals data

        Parameters
        ----------
        query : ast.Select
            Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Hubspot Deals matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition

        """

        select_statement_parser = SELECTQueryParser(
            query,
            "deals",
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        deals_df = pd.json_normalize(self.get_deals(limit=result_limit))
        select_statement_executor = SELECTQueryExecutor(
            deals_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        deals_df = select_statement_executor.execute_query()

        return deals_df

    def insert(self, query: ast.Insert) -> None:
        """
        Inserts data into HubSpot "POST /crm/v3/objects/deals/batch/create" API endpoint.

        Parameters
        ----------
        query : ast.Insert
           Given SQL INSERT query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        insert_statement_parser = INSERTQueryParser(
            query,
            supported_columns=['amount', 'dealname', 'pipeline', 'closedate', 'dealstage', 'hubspot_owner_id'],
            mandatory_columns=['dealname'],
            all_mandatory=False,
        )
        deals_data = insert_statement_parser.parse_query()
        self.create_deals(deals_data)

    def update(self, query: ast.Update) -> None:
        """
        Updates data from HubSpot "PATCH /crm/v3/objects/deals/batch/update" API endpoint.

        Parameters
        ----------
        query : ast.Update
           Given SQL UPDATE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        update_statement_parser = UPDATEQueryParser(query)
        values_to_update, where_conditions = update_statement_parser.parse_query()

        deals_df = pd.json_normalize(self.get_deals())
        update_query_executor = UPDATEQueryExecutor(
            deals_df,
            where_conditions
        )

        deals_df = update_query_executor.execute_query()
        deal_ids = deals_df['id'].tolist()
        self.update_deals(deal_ids, values_to_update)

    def delete(self, query: ast.Delete) -> None:
        """
        Deletes data from HubSpot "DELETE /crm/v3/objects/deals/batch/archive" API endpoint.

        Parameters
        ----------
        query : ast.Delete
           Given SQL DELETE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        delete_statement_parser = DELETEQueryParser(query)
        where_conditions = delete_statement_parser.parse_query()

        deals_df = pd.json_normalize(self.get_deals())
        delete_query_executor = DELETEQueryExecutor(
            deals_df,
            where_conditions
        )

        deals_df = delete_query_executor.execute_query()
        deal_ids = deals_df['id'].tolist()
        self.delete_deals(deal_ids)

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_deals(limit=1)).columns.tolist()

    def get_deals(self, **kwargs) -> List[Dict]:
        hubspot = self.handler.connect()
        deals = hubspot.crm.deals.get_all(**kwargs)
        deals_dict = [
            {
                "id": deal.id,
                "dealname": deal.properties["dealname"],
                "amount": deal.properties.get("amount", None),
                "pipeline": deal.properties.get("pipeline", None),
                "closedate": deal.properties.get("closedate", None),
                "dealstage": deal.properties.get("dealstage", None),
                "hubspot_owner_id": deal.properties.get("hubspot_owner_id", None),
                "createdate": deal.properties["createdate"],
                "hs_lastmodifieddate": deal.properties["hs_lastmodifieddate"],
            }
            for deal in deals
        ]
        return deals_dict

    def create_deals(self, deals_data: List[Dict[Text, Any]]) -> None:
        hubspot = self.handler.connect()
        deals_to_create = [HubSpotObjectInputCreate(properties=deal) for deal in deals_data]
        try:
            created_deals = hubspot.crm.deals.batch_api.create(
                HubSpotBatchObjectBatchInput(inputs=deals_to_create),
            )
            logger.info(f"Deals created with ID's {[created_deal.id for created_deal in created_deals.results]}")
        except Exception as e:
            raise Exception(f"Deals creation failed {e}")

    def update_deals(self, deal_ids: List[Text], values_to_update: Dict[Text, Any]) -> None:
        hubspot = self.handler.connect()
        deals_to_update = [HubSpotObjectBatchInput(id=deal_id, properties=values_to_update) for deal_id in deal_ids]
        try:
            updated_deals = hubspot.crm.deals.batch_api.update(
                HubSpotBatchObjectBatchInput(inputs=deals_to_update),
            )
            logger.info(f"Deals with ID {[updated_deal.id for updated_deal in updated_deals.results]} updated")
        except Exception as e:
            raise Exception(f"Deals update failed {e}")

    def delete_deals(self, deal_ids: List[Text]) -> None:
        hubspot = self.handler.connect()
        deals_to_delete = [HubSpotObjectId(id=deal_id) for deal_id in deal_ids]
        try:
            hubspot.crm.deals.batch_api.archive(
                HubSpotBatchObjectIdInput(inputs=deals_to_delete),
            )
            logger.info("Deals deleted")
        except Exception as e:
            raise Exception(f"Deals deletion failed {e}")



from typing import Dict, Optional

import pandas as pd

from mindsdb.integrations.handlers.rag_handler.ingest import RAGIngestor
from mindsdb.integrations.handlers.rag_handler.rag import RAGQuestionAnswerer
from mindsdb.integrations.handlers.rag_handler.settings import (
    DEFAULT_EMBEDDINGS_MODEL,
    RAGHandlerParameters,
    build_llm_params,
)
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities import log


logger = log.getLogger(__name__)

logger.warning("\nThe RAG handler has been deprecated and is no longer being actively supported. \n"
               "It will be fully removed in v24.8.x.x, "
               "for RAG workflows, please migrate to "
               "Agents + Retrieval skill. \n"
               "Example usage can be found here: \n"
               "https://github.com/mindsdb/mindsdb_python_sdk/blob/staging/examples"
               "/using_agents_with_retrieval.py")


class RAGHandler(BaseMLEngine):
    """
    RAGHandler is a MindsDB integration with supported LLM APIs allows users to run question answering
    on their data by providing a question.

    The User is able to provide data that provides context for the questions, see create() method for more details.

    """

    name = "rag"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generative = True

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        if "using" not in args:
            raise Exception(
                "RAG engine requires a USING clause! Refer to its documentation for more details."
            )

    def create(
            self,
            target: str,
            df: pd.DataFrame = None,
            args: Optional[Dict] = None,
    ):
        """
        Dispatch is running embeddings and storing in a VectorDB, unless user already has embeddings persisted
        """
        # get api key from user input on create ML_ENGINE or create MODEL
        args = args["using"]

        ml_engine_args = self.engine_storage.get_connection_args()

        # for a model created with USING, only get api for that specific llm type
        args.update({k:v for k, v in ml_engine_args.items() if args["llm_type"] in k})

        input_args = build_llm_params(args)

        args = RAGHandlerParameters(**input_args)

        # create folder for vector store to persist embeddings or load from existing folder
        args.vector_store_storage_path = self.engine_storage.folder_get(
            args.vector_store_folder_name
        )

        if args.run_embeddings:
            if "context_columns" not in args and df is not None:
                # if no context columns provided, use all columns in df
                logger.info("No context columns provided, using all columns in df")
                args.context_columns = df.columns.tolist()

            if "embeddings_model_name" not in args:
                logger.info(
                    f"No embeddings model provided in query, using default model: {DEFAULT_EMBEDDINGS_MODEL}"
                )

            if df is not None or args.url is not None:
                # if user provides a dataframe or url, run embeddings and store in vector store

                ingestor = RAGIngestor(args=args, df=df)
                ingestor.embeddings_to_vectordb()

        else:
            # Note this should only be run if run_embeddings is false or if no data is provided in query
            logger.info("Skipping embeddings and ingestion into Chroma VectorDB")

        export_args = args.dict(exclude={"llm_params"})
        # 'callbacks' aren't json serializable, we do this to avoid errors
        export_args["llm_params"] = args.llm_params.dict(exclude={"callbacks"})

        # for mindsdb cloud, store data in shared file system
        # for cloud version of mindsdb to make it be usable by all mindsdb nodes
        self.engine_storage.folder_sync(args.vector_store_folder_name)

        self.model_storage.json_set("args", export_args)

    def update(self, args) -> None:

        # build llm params from user input args in update query
        updated_args = build_llm_params(args["using"], update=True)

        # get current model args
        current_model_args = self.model_storage.json_get("args")["using"]

        # update current args with new args
        current_model_args.update(updated_args)

        # validate updated args are valid
        RAGHandlerParameters(**build_llm_params(current_model_args))

        # if valid, update model args
        self.model_storage.json_set("args", current_model_args)

    def predict(self, df: pd.DataFrame = None, args: dict = None):
        """
        Dispatch is performed depending on the underlying model type. Currently, only question answering
        is supported.
        """

        input_args = build_llm_params(self.model_storage.json_get("args"))

        args = RAGHandlerParameters(**input_args)

        args.vector_store_storage_path = self.engine_storage.folder_get(
            args.vector_store_folder_name
        )

        # get question answering results
        question_answerer = RAGQuestionAnswerer(args=args)

        # get question from sql query
        # e.g. where question = 'What is the capital of France?'
        response = question_answerer(df[args.input_column].tolist()[0])

        return pd.DataFrame(response)



import json

import pandas as pd
from pandas import DataFrame
from google.auth.transport.requests import Request
from google.oauth2 import service_account
from google.oauth2.credentials import Credentials
from googleapiclient.discovery import build
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE
from .google_content_shopping_tables import AccountsTable, OrdersTable, ProductsTable
from mindsdb.integrations.libs.api_handler import APIHandler, FuncParser
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
)
from mindsdb.utilities import log

logger = log.getLogger(__name__)

class GoogleContentShoppingHandler(APIHandler):
    """
        A class for handling connections and interactions with the Google Content API for Shopping.
    """
    name = 'google_content_shopping'

    def __init__(self, name: str, **kwargs):
        """
        Initialize the Google Content API for Shopping handler.
        Args:
            name (str): name of the handler
            kwargs (dict): additional arguments
        """
        super().__init__(name)
        self.token = None
        self.service = None
        self.connection_data = kwargs.get('connection_data', {})
        self.fs_storage = kwargs['file_storage']
        self.credentials_file = self.connection_data.get('credentials', None)
        self.merchant_id = self.connection_data.get('merchant_id', None)
        self.credentials = None
        self.scopes = ['https://www.googleapis.com/auth/content']
        self.is_connected = False
        accounts = AccountsTable(self)
        self.accounts = accounts
        self._register_table('Accounts', accounts)
        orders = OrdersTable(self)
        self.orders = orders
        self._register_table('Orders', orders)
        products = ProductsTable(self)
        self.products = products
        self._register_table('Products', products)

    def connect(self):
        """
        Set up any connections required by the handler
        Should return output of check_connection() method after attempting
        connection. Should switch self.is_connected.
        Returns:
            HandlerStatusResponse
        """
        if self.is_connected is True:
            return self.service
        if self.credentials_file:
            try:
                json_str_bytes = self.fs_storage.file_get('token_content.json')
                json_str = json_str_bytes.decode()
                self.credentials = Credentials.from_authorized_user_info(info=json.loads(json_str), scopes=self.scopes)
            except Exception:
                self.credentials = None
            if not self.credentials or not self.credentials.valid:
                if self.credentials and self.credentials.expired and self.credentials.refresh_token:
                    self.credentials.refresh(Request())
                else:
                    self.credentials = service_account.Credentials.from_service_account_file(
                        self.credentials_file, scopes=self.scopes)
            # Save the credentials for the next run
            json_str = self.credentials.to_json()
            self.fs_storage.file_set('token_content.json', json_str.encode())
            self.service = build('content', 'v2.1', credentials=self.credentials)
        return self.service

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler
        Returns:
            HandlerStatusResponse
        """
        response = StatusResponse(False)

        try:
            service = self.connect()
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to Google Content API for Shopping: {e}!')
            response.error_message = e

        self.is_connected = response.success
        return response

    def native_query(self, query: str = None) -> Response:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (Any): query in native format (str for sql databases,
                dict for mongo, api's json etc)
        Returns:
            HandlerResponse
        """
        method_name, params = FuncParser().from_string(query)

        df = self.call_application_api(method_name, params)

        return Response(
            RESPONSE_TYPE.TABLE,
            data_frame=df
        )

    def get_accounts(self, params: dict = None) -> DataFrame:
        """
        Get accounts
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        page_token = None
        accounts = pd.DataFrame(columns=self.accounts.get_columns())
        if params['account_id']:
            result = service.accounts().get(merchantId=self.merchant_id, accountId=params['account_id']).execute()
            accounts = pd.DataFrame(result, columns=self.accounts.get_columns())
            return accounts
        while True:
            result = service.accounts().list(merchantId=self.merchant_id, page_token=page_token, **params).execute()
            accounts = pd.concat(
                [accounts, pd.DataFrame(result['resources'], columns=self.accounts.get_columns())],
                ignore_index=True
            )
            page_token = result.get('nextPageToken')
            if not page_token:
                break

        if params['startId'] and params['endId']:
            start_id = int(params['startId'])
            end_id = int(params['endId'])
        elif params['startId']:
            start_id = int(params['startId'])
            end_id = start_id + 10
        elif params['endId']:
            end_id = int(params['endId'])
            start_id = end_id - 10
        else:
            raise Exception('startId or endId must be specified')

        accounts = accounts.drop(accounts[(accounts['id'] < start_id) | (accounts['id'] > end_id)].index)

        return accounts

    def delete_accounts(self, params: dict = None) -> DataFrame:
        """
        Delete accounts
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        args = {}
        if params['force']:
            args = {'force': params['force']}
        if params['accountId']:
            result = service.accounts().delete(merchantId=self.merchant_id, accountId=params['accountId'],
                                               **args).execute()
            return result
        else:
            df = pd.DataFrame(columns=['accountId', 'status'])
            if not params['startId']:
                start_id = int(params['endId']) - 10
            elif not params['endId']:
                end_id = int(params['startId']) + 10
            else:
                start_id = int(params['startId'])
                end_id = int(params['endId'])

            for i in range(start_id, end_id):
                service.accounts().delete(merchantId=self.merchant_id, accountId=i, **args).execute()
                df = pd.concat([df, pd.DataFrame([{'accountId': str(i), 'status': 'deleted'}])], ignore_index=True)
            return df

    def get_orders(self, params: dict = None) -> DataFrame:
        """
        Get orders
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        page_token = None
        orders = pd.DataFrame(columns=self.orders.get_columns())
        args = {
            key: value for key, value in params.items() if key in ['maxResults',
                                                                   'statuses',
                                                                   'acknowledged',
                                                                   'placedDateStart',
                                                                   'placedDateEnd',
                                                                   'orderBy']
                                                           and value is not None
        }
        if params['order_id']:
            result = service.orders().get(merchantId=self.merchant_id, orderId=params['order_id'], **args).execute()
            orders = pd.DataFrame(result, columns=self.orders.get_columns())
            return orders
        while True:
            result = service.orders().list(merchantId=self.merchant_id, page_token=page_token, **args).execute()
            orders = pd.concat(
                [orders, pd.DataFrame(result['resources'], columns=self.orders.get_columns())],
                ignore_index=True
            )
            page_token = result.get('nextPageToken')
            if not page_token:
                break

        if params['startId'] and params['endId']:
            start_id = int(params['startId'])
            end_id = int(params['endId'])
        elif params['startId']:
            start_id = int(params['startId'])
            end_id = start_id + 10
        elif params['endId']:
            end_id = int(params['endId'])
            start_id = end_id - 10
        else:
            raise Exception('startId or endId must be specified')

        orders = orders.drop(orders[(orders['id'] < start_id) | (orders['id'] > end_id)].index)

        return orders

    def delete_orders(self, params: dict = None) -> DataFrame:
        """
        Delete orders
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        if params['order_id']:
            result = service.orders().delete(merchantId=self.merchant_id, orderId=params['order_id']).execute()
            return result
        else:
            df = pd.DataFrame(columns=['orderId', 'status'])
            if not params['startId']:
                start_id = int(params['endId']) - 10
            elif not params['endId']:
                end_id = int(params['startId']) + 10
            else:
                start_id = int(params['startId'])
                end_id = int(params['endId'])

            for i in range(start_id, end_id):
                service.orders().delete(merchantId=self.merchant_id, orderId=i).execute()
                df = pd.concat([df, pd.DataFrame([{'orderId': str(i), 'status': 'deleted'}])], ignore_index=True)
            return df

    def get_products(self, params: dict = None) -> DataFrame:
        """
        Get products
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        page_token = None
        products = pd.DataFrame(columns=self.products.get_columns())
        if params['product_id']:
            result = service.products().get(merchantId=self.merchant_id, productId=params['product_id']).execute()
            products = pd.DataFrame(result, columns=self.products.get_columns())
            return products
        while True:
            result = service.products().list(merchantId=self.merchant_id, page_token=page_token).execute()
            products = pd.concat(
                [products, pd.DataFrame(result['resources'], columns=self.products.get_columns())],
                ignore_index=True
            )
            page_token = result.get('nextPageToken')
            if not page_token:
                break

        if params['startId'] and params['endId']:
            start_id = int(params['startId'])
            end_id = int(params['endId'])
        elif params['startId']:
            start_id = int(params['startId'])
            end_id = start_id + 10
        elif params['endId']:
            end_id = int(params['endId'])
            start_id = end_id - 10
        else:
            raise Exception('startId or endId must be specified')

        products = products.drop(products[(products['id'] < start_id) | (products['id'] > end_id)].index)

        return products

    def update_products(self, params: dict = None) -> DataFrame:
        """
        Update products
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        body = {
            key: value for key, value in params.items() if key in self.products.get_columns()
        }
        service = self.connect()
        if params['product_id']:
            result = service.products().update(merchantId=self.merchant_id, productId=params['product_id'],
                                               updateMask=params['updateMask'], body=body).execute()

            return result
        else:
            df = pd.DataFrame(columns=['productId', 'status'])
            if not params['startId']:
                start_id = int(params['endId']) - 10
            elif not params['endId']:
                end_id = int(params['startId']) + 10
            else:
                start_id = int(params['startId'])
                end_id = int(params['endId'])

            for i in range(start_id, end_id):
                service.products().update(merchantId=self.merchant_id, productId=i,
                                          updateMask=params['updateMask'], body=body).execute()
                df = pd.concat([df, pd.DataFrame([{'productId': str(i), 'status': 'updated'}])], ignore_index=True)
            return df

    def delete_products(self, params: dict = None) -> DataFrame:
        """
        Delete products
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        args = {
            key: value for key, value in params.items() if key in ['feedId'] and value is not None
        }
        if params['product_id']:
            result = service.products().delete(merchantId=self.merchant_id, productId=params['product_id'],
                                               **args).execute()
            return result
        else:
            df = pd.DataFrame(columns=['productId', 'status'])
            if not params['startId']:
                start_id = int(params['endId']) - 10
            elif not params['endId']:
                end_id = int(params['startId']) + 10
            else:
                start_id = int(params['startId'])
                end_id = int(params['endId'])

            for i in range(start_id, end_id):
                service.products().delete(merchantId=self.merchant_id, productId=i, **args).execute()
                df = pd.concat([df, pd.DataFrame([{'productId': str(i), 'status': 'deleted'}])], ignore_index=True)
            return df

    def call_application_api(self, method_name: str = None, params: dict = None) -> DataFrame:
        """
        Call Google Search API and map the data to pandas DataFrame
        Args:
            method_name (str): method name
            params (dict): query parameters
        Returns:
            DataFrame
        """
        if method_name == 'get_accounts':
            return self.get_accounts(params)
        elif method_name == 'delete_accounts':
            return self.delete_accounts(params)
        elif method_name == 'get_orders':
            return self.get_orders(params)
        elif method_name == 'delete_orders':
            return self.delete_orders(params)
        elif method_name == 'get_products':
            return self.get_products(params)
        elif method_name == 'update_products':
            return self.update_products(params)
        elif method_name == 'delete_products':
            return self.delete_products(params)
        else:
            raise NotImplementedError(f'Unknown method {method_name}')



import mindsdb.interfaces.storage.db as db
from mindsdb.api.mysql.mysql_proxy.mysql_proxy import MysqlProxy
from mindsdb.utilities import log
from mindsdb.utilities.functions import init_lexer_parsers


def start(verbose=False):
    logger = log.getLogger(__name__)
    logger.info("MySQL API is starting..")
    db.init()
    init_lexer_parsers()

    MysqlProxy.startProxy()



from typing import Callable

from mindsdb.interfaces.chatbot.chatbot_message import ChatBotMessage
from mindsdb.interfaces.chatbot.chatbot_response import ChatBotResponse


class RealtimeChatHandler:
    """Interface to send and receive messages over a chat application (Slack, RocketChat, etc)"""

    def __init__(self, name: str, on_message: Callable[[ChatBotMessage], None]):
        self.name = name
        # Should be called every time a message is received.
        self.on_message = on_message

    def connect(self):
        """Connects to chat application and starts listening for messages."""
        raise NotImplementedError()

    def disconnect(self):
        """Disconnects from the chat application and stops listening for messages."""
        raise NotImplementedError()

    def send_message(self, message: ChatBotMessage) -> ChatBotResponse:
        """
        Sends a message through the chat application

        Parameters:
            message (ChatBotMessage): Message to send

        Returns:
            response (ChatBotResponse): Response indicating whether the message was sent successfully
        """
        raise NotImplementedError()



from mindsdb.integrations.libs.const import HANDLER_TYPE

from mindsdb.integrations.handlers.autokeras_handler.__about__ import (
    __version__ as version,
    __description__ as description,
)

try:
    from .autokeras_handler import AutokerasHandler as Handler

    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = "AutoKeras"
name = "autokeras"
type = HANDLER_TYPE.ML
icon_path = 'icon.svg'
permanent = False

__all__ = ["Handler", "version", "name", "type", "title", "description", "import_error", "icon_path"]



import unittest
from mindsdb.integrations.handlers.libsql_handler.libsql_handler import LibSQLHandler
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE


class LibSQLHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "database": "tests/test.db",
        }
        cls.handler = LibSQLHandler("test_libsql_handler", cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_create_table(self):
        query = "CREATE TABLE IF NOT EXISTS user (id INTEGER PRIMARY KEY, name TEXT, age INTEGER)"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.OK

    def test_2_insert_rows(self):
        query = (
            "INSERT OR IGNORE INTO user (name, age) VALUES ('Alice', 30), ('Bob', 25)"
        )
        result = self.handler.native_query(query)
        print(result)
        assert result.type is RESPONSE_TYPE.OK

    def test_3_native_query_select(self):
        query = "SELECT * FROM user"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_4_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not RESPONSE_TYPE.ERROR

    def test_5_get_columns(self):
        columns = self.handler.get_columns("user")
        assert columns.type is not RESPONSE_TYPE.ERROR


if __name__ == "__main__":
    unittest.main()



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Amazon Aurora DB cluster.'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password to authenticate the user with the Amazon Aurora DB cluster.',
        'secret': True
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'The database name to use when connecting with the Amazon Aurora DB cluster.'
    },
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the Amazon Aurora DB cluster.'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'The TCP/IP port of the Amazon Aurora DB cluster. Must be an integer.'
    },
    db_engine={
        'type': ARG_TYPE.STR,
        'description': "The database engine of the Amazon Aurora DB cluster. This can take one of two values: 'mysql' or 'postgresql'. This parameter is optional, but if it is not provided, 'aws_access_key_id' and 'aws_secret_access_key' parameters must be provided"
    },
    aws_access_key_id={
        'type': ARG_TYPE.STR,
        'description': "The access key for the AWS account. This parameter is optional and is only required to be provided if the 'db_engine' parameter is not provided."
    },
    aws_secret_access_key={
        'type': ARG_TYPE.STR,
        'description': "The secret key for the AWS account. This parameter is optional and is only required to be provided if the 'db_engine' parameter is not provided.",
        'secret': True
    },
)

connection_args_example = OrderedDict(
    db_engine='mysql',
    host='mysqlcluster.cluster-123456789012.us-east-1.rds.amazonaws.com',
    port=3306,
    user='root',
    password='password',
    database='database'
)



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    credentials={
        'type': ARG_TYPE.PATH,
        'description': 'The path to the credentials file. If not specified, the default credentials are used.'
    }
)

connection_args_example = OrderedDict(
    credentials='/path/to/credentials.json'
)



from typing import List, Optional

import json
from langchain_core.callbacks.manager import CallbackManagerForRetrieverRun
from langchain_core.documents import Document
from langchain_core.language_models import BaseChatModel
from langchain_core.retrievers import BaseRetriever
from pydantic import Field, PrivateAttr

from mindsdb.integrations.utilities.rag.settings import (
    RAGPipelineModel,
    DEFAULT_QUESTION_REFORMULATION_TEMPLATE
)
from mindsdb.integrations.utilities.rag.retrievers.retriever_factory import create_retriever


class MultiHopRetriever(BaseRetriever):
    """A retriever that implements multi-hop question reformulation strategy.

    This retriever takes a base retriever and uses an LLM to generate follow-up
    questions based on the initial results. It then retrieves documents for each
    follow-up question and combines all results.
    """

    base_retriever: BaseRetriever = Field(description="Base retriever to use for document lookup")
    llm: BaseChatModel = Field(description="LLM to use for generating follow-up questions")
    max_hops: int = Field(default=3, description="Maximum number of follow-up questions to generate")
    reformulation_template: str = Field(
        default=DEFAULT_QUESTION_REFORMULATION_TEMPLATE,
        description="Template for reformulating questions"
    )

    _asked_questions: set = PrivateAttr(default_factory=set)

    @classmethod
    def from_config(cls, config: RAGPipelineModel) -> "MultiHopRetriever":
        """Create a MultiHopRetriever from a RAGPipelineModel config."""
        if config.multi_hop_config is None:
            raise ValueError("multi_hop_config must be set for MultiHopRetriever")

        # Create base retriever based on type
        base_retriever = create_retriever(config, config.multi_hop_config.base_retriever_type)

        return cls(
            base_retriever=base_retriever,
            llm=config.llm,
            max_hops=config.multi_hop_config.max_hops,
            reformulation_template=config.multi_hop_config.reformulation_template
        )

    def _get_relevant_documents(
        self, query: str, *, run_manager: Optional[CallbackManagerForRetrieverRun] = None
    ) -> List[Document]:
        """Get relevant documents using multi-hop retrieval."""
        if query in self._asked_questions:
            return []

        self._asked_questions.add(query)

        # Get initial documents
        docs = self.base_retriever._get_relevant_documents(query)
        if not docs or len(self._asked_questions) >= self.max_hops:
            return docs

        # Generate follow-up questions
        context = "\n".join(doc.page_content for doc in docs)
        prompt = self.reformulation_template.format(
            question=query,
            context=context
        )

        try:
            follow_up_questions = json.loads(self.llm.invoke(prompt))
            if not isinstance(follow_up_questions, list):
                return docs
        except (json.JSONDecodeError, TypeError):
            return docs

        # Get documents for follow-up questions
        for question in follow_up_questions:
            if isinstance(question, str):
                follow_up_docs = self._get_relevant_documents(question)
                docs.extend(follow_up_docs)

        return docs



from stability_sdk import client
from stability_sdk.client import generation
from PIL import Image
import io
import requests


class StabilityAPIClient:

    def _init_(self, api_key, dir_to_save, engine="stable-diffusion-xl-1024-v1-0", upscale_engine="esrgan-v1-x2plus"):
        """Initialize the stability wrapper api client.

        Args:
            api_key: Stability AI API Key
            dir_to_save: The local directory path to save the response images from the API
            engine (str, optional): Stability AI engine to use. Defaults to "stable-diffusion-xl-1024-v1-0".
            upscale_engine (str, optional): Stability AI upscaling engine to use. Defaults to "esrgan-v1-x2plus".

        Raises:
            ValueError: For unknown engine or upscale engine
        """
        self.api_key = api_key
        self.STABILITY_HOST = 'grpc.stability.ai:443'
        self.save_dir = dir_to_save + "/" if not dir_to_save.endswith("/") else dir_to_save
        self.available_engines = self.get_existing_engines()
        if not self._is_valid_engine(engine):
            raise ValueError("Unknown engine. The available engines are - " + self.available_engines)
        if not self._is_valid_engine(upscale_engine):
            raise ValueError("Unknown upscale engine. The available engines are - " + self.available_engines)

        self.stability_api = client.StabilityInference(host=self.STABILITY_HOST,
                                                       key=self.api_key,
                                                       engine=engine,
                                                       upscale_engine=upscale_engine,
                                                       verbose=True)

    def save_image(self, artifact):
        """Save the binary image in the artifact to the local directory

        Args:
            artifact: Artifact returned by the API

        Returns:
            The local image path.
        """
        img = Image.open(io.BytesIO(artifact.binary))
        path = self.save_dir + str(artifact.seed) + ".png"
        img.save(path)
        return path

    def _process_artifacts(self, artifacts):
        """Process the artificats returned by the API

        Args:
            artifacts :Artifact returned by the API

        Returns:
            The saved image paths
        """
        saved_image_paths = []
        for resp in artifacts:
            for artifact in resp.artifacts:
                if artifact.finish_reason == generation.FILTER:
                    saved_image_paths.append(
                        "Your request activated the API's safety filters \
                        and could not be processed. Please modify the prompt and try again.")
                if artifact.type == generation.ARTIFACT_IMAGE:
                    saved_image_paths.append(self.save_image(artifact))
        return saved_image_paths

    def get_existing_engines(self):
        """Get the existing engines

        Returns: Engines supported by the API
        """
        url = "https://api.stability.ai/v1/engines/list"
        if self.api_key is None:
            raise Exception("Missing Stability API key.")

        response = requests.get(url, headers={
            "Authorization": f"Bearer {self.api_key}"
        })

        if response.status_code != 200:
            raise Exception("Non-200 response: " + str(response.text))

        payload = response.json()

        return {engine["id"]: engine for engine in payload}

    def _is_valid_engine(self, engine_id):
        """Validates the given engine id against the supported engines by Stability

        Args:
            engine_id: The engine id to check

        Returns:
            True if valid engine else False
        """
        return engine_id in self.available_engines

    def _read_image_url(self, image_url):
        """Downloads the given image url.

        Args:
            image_url: The image url to download

        Returns:
            Downloaded image
        """
        return Image.open(requests.get(image_url, stream=True).raw)

    def text_to_image(self, prompt, height=1024, width=1024):
        """Converts the given text to image using stability API.

        Args:
            prompt: The given text
            height (int, optional): Height of the image. Defaults to 1024.
            width (int, optional): Width of the image. Defaults to 1024.

        Returns:
            The local saved paths of the generated images
        """
        answers = self.stability_api.generate(
            prompt=prompt,
            height=height,
            width=width
        )

        saved_images = self._process_artifacts(answers)

        return saved_images

    def image_to_image(self, image_url, prompt=None,
                       height=1024, width=1024,
                       mask_image_url=None):
        """Image to Image inpainting + masking

        Args:
            image_url: The image url
            prompt: The given text
            height (int, optional): Height of the image. Defaults to 1024.
            width (int, optional): Width of the image. Defaults to 1024.
            mask_image_url (string, optional): The mask image url for masking. Defaults to None.

        Returns:
            The local saved paths of the generated images
        """
        img = self._read_image_url(image_url)
        mask_img = None if (mask_image_url is None) else self._read_image_url(mask_image_url)
        if prompt is None:
            prompt = ""
        answers = self.stability_api.generate(prompt=prompt, init_image=img, mask_image=mask_img, width=width, height=height)

        saved_images = self._process_artifacts(answers)

        return saved_images

    def image_upscaling(self, image_url, height=None, width=None, prompt=None):
        """Image upscaling

        Args:
            image_url: The image url
            height (int, optional): Height of the image. Defaults to None.
            width (int, optional): Width of the image. Defaults to None.
            prompt: The given text

        Returns:
            The local saved paths of the generated images
        """
        img = self._read_image_url(image_url)
        if height is not None and width is not None:
            raise Exception("Either height or width can be given. Refer - https://platform.stability.ai/docs/features/image-upscaling#initial-generation-parameters")

        if height is None and width is None:
            answers = self.stability_api.upscale(init_image=img, prompt=prompt)

        if height is not None:
            answers = self.stability_api.upscale(init_image=img, height=height, prompt=prompt)

        if width is not None:
            answers = self.stability_api.upscale(init_image=img, width=width, prompt=prompt)

        saved_images = self._process_artifacts(answers)

        return saved_images



"""predictor_training_metadata

Revision ID: f6dc924079fa
Revises: a8a3fac369e7
Create Date: 2024-11-29 15:06:47.269229

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa

# revision identifiers, used by Alembic.
revision = 'f6dc924079fa'
down_revision = 'a8a3fac369e7'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.add_column(sa.Column('training_metadata', sa.JSON(), nullable=True))
        batch_op.drop_column('is_custom')
        batch_op.drop_column('hostname')

    predictor_table = sa.Table(
        'predictor',
        sa.MetaData(),
        sa.Column('id', sa.Integer()),
        sa.Column('training_metadata', sa.JSON())
    )

    op.execute(predictor_table.update().values(training_metadata={}))
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.alter_column(
            'training_metadata',
            nullable=False
        )


def downgrade():
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.add_column(sa.Column('hostname', sa.VARCHAR(), nullable=True))
        batch_op.add_column(sa.Column('is_custom', sa.BOOLEAN(), nullable=True))
        batch_op.drop_column('training_metadata')



from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import DEFAULT_CAPABILITIES


class ServerCapabilities():
    def __init__(self, capabilities):
        self._capabilities = capabilities

    def has(self, cap):
        return cap & self._capabilities > 0

    def set(self, cap, value=True):
        if value:
            self._capabilities = self._capabilities | cap
        else:
            self._capabilities = self._capabilities & (~cap)

    @property
    def value(self):
        return self._capabilities


server_capabilities = ServerCapabilities(DEFAULT_CAPABILITIES)



import traceback
import json
import csv
from io import BytesIO, StringIO
from pathlib import Path
import codecs

import filetype
import pandas as pd
from charset_normalizer import from_bytes
from langchain_text_splitters import RecursiveCharacterTextSplitter

from mindsdb.utilities import log

logger = log.getLogger(__name__)

DEFAULT_CHUNK_SIZE = 500
DEFAULT_CHUNK_OVERLAP = 250


class FileDetectError(Exception):
    ...


def decode(file_obj: BytesIO) -> StringIO:
    byte_str = file_obj.read()
    # Move it to StringIO
    try:
        # Handle Microsoft's BOM "special" UTF-8 encoding
        if byte_str.startswith(codecs.BOM_UTF8):
            data_str = StringIO(byte_str.decode("utf-8-sig"))
        else:
            file_encoding_meta = from_bytes(
                byte_str[: 32 * 1024],
                steps=32,  # Number of steps/block to extract from my_byte_str
                chunk_size=1024,  # Set block size of each extraction)
                explain=False,
            )
            best_meta = file_encoding_meta.best()
            errors = "strict"
            if best_meta is not None:
                encoding = file_encoding_meta.best().encoding

                try:
                    data_str = StringIO(byte_str.decode(encoding, errors))
                except UnicodeDecodeError:
                    encoding = "utf-8"
                    errors = "replace"

                    data_str = StringIO(byte_str.decode(encoding, errors))
            else:
                encoding = "utf-8"
                errors = "replace"

                data_str = StringIO(byte_str.decode(encoding, errors))
    except Exception as e:
        logger.error(traceback.format_exc())
        raise FileDetectError("Could not load into string") from e

    return data_str


class FormatDetector:

    def get(self, name, file_obj: BytesIO = None):
        format = self.get_format_by_name(name)
        if format is None and file_obj is not None:
            format = self.get_format_by_content(file_obj)

        if format is not None:
            return format
        raise FileDetectError(f'Unable to detect format: {name}')

    def get_format_by_name(self, filename):
        extension = Path(filename).suffix.strip(".").lower()
        if extension == "tsv":
            extension = "csv"
        return extension or None

    def get_format_by_content(self, file_obj):
        if self.is_parquet(file_obj):
            return "parquet"

        file_type = filetype.guess(file_obj)
        if file_type is None:
            return

        if file_type.mime in {
            "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
            "application/vnd.ms-excel",
        }:
            return 'xlsx'

        if file_type.mime == 'application/pdf':
            return "pdf"

        file_obj = decode(file_obj)

        if self.is_json(file_obj):
            return "json"

        if self.is_csv(file_obj):
            return "csv"

    def is_json(self, data_obj: StringIO) -> bool:
        # see if its JSON
        text = data_obj.read(100).strip()
        data_obj.seek(0)
        if len(text) > 0:
            # it looks like a json, then try to parse it
            if text.startswith("{") or text.startswith("["):
                try:
                    json.loads(data_obj.read())
                    return True
                except Exception:
                    return False
                finally:
                    data_obj.seek(0)
        return False

    def is_csv(self, data_obj: StringIO) -> bool:
        sample = data_obj.readline()  # trying to get dialect from header
        data_obj.seek(0)
        try:
            csv.Sniffer().sniff(sample)

        except Exception:
            return False

    def is_parquet(self, data: BytesIO) -> bool:
        # Check first and last 4 bytes equal to PAR1.
        # Refer: https://parquet.apache.org/docs/file-format/
        parquet_sig = b"PAR1"
        data.seek(0, 0)
        start_meta = data.read(4)
        data.seek(-4, 2)
        end_meta = data.read()
        data.seek(0)
        if start_meta == parquet_sig and end_meta == parquet_sig:
            return True
        return False


class FileReader:

    def _get_csv_dialect(self, buffer) -> csv.Dialect:
        sample = buffer.readline()  # trying to get dialect from header
        buffer.seek(0)
        try:
            if isinstance(sample, bytes):
                sample = sample.decode()
            accepted_csv_delimiters = [",", "\t", ";"]
            try:
                dialect = csv.Sniffer().sniff(
                    sample, delimiters=accepted_csv_delimiters
                )
                dialect.doublequote = (
                    True  # assume that all csvs have " as string escape
                )
            except Exception:
                dialect = csv.reader(sample).dialect
                if dialect.delimiter not in accepted_csv_delimiters:
                    raise Exception(
                        f"CSV delimeter '{dialect.delimiter}' is not supported"
                    )

        except csv.Error:
            dialect = None
        return dialect

    def read(self, format, file_obj: BytesIO, **kwargs) -> pd.DataFrame:
        func = {
            'parquet': self.read_parquet,
            'csv': self.read_csv,
            'xlsx': self.read_excel,
            'pdf': self.read_pdf,
            'json': self.read_json,
            'txt': self.read_txt,
        }

        if format not in func:
            raise FileDetectError(f'Unsupported format: {format}')
        func = func[format]

        return func(file_obj, **kwargs)

    def read_csv(self, file_obj: BytesIO, **kwargs):
        file_obj = decode(file_obj)
        dialect = self._get_csv_dialect(file_obj)

        return pd.read_csv(file_obj, sep=dialect.delimiter, index_col=False)

    def read_txt(self, file_obj: BytesIO, **kwargs):
        file_obj = decode(file_obj)

        try:
            from langchain_core.documents import Document
        except ImportError:
            raise ImportError(
                "To import TXT document please install 'langchain-community':\n"
                "    pip install langchain-community"
            )
        text = file_obj.read()

        file_name = None
        if hasattr(file_obj, "name"):
            file_name = file_obj.name
        metadata = {"source": file_name}
        documents = [Document(page_content=text, metadata=metadata)]

        text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=DEFAULT_CHUNK_SIZE, chunk_overlap=DEFAULT_CHUNK_OVERLAP
        )

        docs = text_splitter.split_documents(documents)
        return pd.DataFrame(
            [
                {"content": doc.page_content, "metadata": doc.metadata}
                for doc in docs
            ]
        )

    def read_pdf(self, file_obj: BytesIO, **kwargs):
        import fitz  # pymupdf

        with fitz.open(stream=file_obj) as pdf:  # open pdf
            text = chr(12).join([page.get_text() for page in pdf])

        text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=DEFAULT_CHUNK_SIZE, chunk_overlap=DEFAULT_CHUNK_OVERLAP
        )

        split_text = text_splitter.split_text(text)

        return pd.DataFrame(
            {"content": split_text, "metadata": [{}] * len(split_text)}
        )

    def read_json(self, file_obj: BytesIO, **kwargs):
        file_obj = decode(file_obj)
        file_obj.seek(0)
        json_doc = json.loads(file_obj.read())
        return pd.json_normalize(json_doc, max_level=0)

    def read_parquet(self, file_obj: BytesIO, **kwargs):
        return pd.read_parquet(file_obj)

    def read_excel(self, file_obj: BytesIO, sheet_name=None, **kwargs) -> pd.DataFrame:

        file_obj.seek(0)
        with pd.ExcelFile(file_obj) as xls:
            if sheet_name is None:
                # No sheet specified: Return list of sheets
                sheet_list = xls.sheet_names
                return pd.DataFrame(sheet_list, columns=["Sheet_Name"])
            else:
                # Specific sheet requested: Load that sheet
                return pd.read_excel(xls, sheet_name=sheet_name)
